import numpy as np

from ..cfdmimplementation import implementation
from .abstract import ReadWrite
from .netcdf import NetCDFWrite


class write(ReadWrite):
    """Write field and domain constructs to a netCDF file.

    **File format**

    See the *fmt* parameter for details on which output netCDF file
    formats are supported.

    **NetCDF variable and dimension names**

    These names are stored within constructs read a from dataset, or
    may be set manually. They are used when writing a field construct
    to the file. If a name has not been set then one will be
    constructed (usually based on the standard name if it exists). The
    names may be modified internally to prevent duplication in the
    file.

    Each construct, or construct component, that corresponds to a
    netCDF variable has the following methods to get, set and remove a
    netCDF variable name: `!nc_get_variable`, `!nc_set_variable` and
    `!nc_del_variable` method

    The domain axis construct has the following methods to get, set
    and remove a netCDF dimension name:
    `~{{package}}.DomainAxis.nc_get_dimension`,
    `~{{package}}.DomainAxis.nc_set_dimension` and
    `~{{package}}.DomainAxis.nc_del_dimension`.

    **NetCDF attributes**

    Field construct properties may be written as netCDF global
    attributes and/or netCDF data variable attributes. See the
    *file_descriptors*, *global_attributes* and *variable_attributes*
    parameters for details.

    **External variables**

    Metadata constructs marked as external are omitted from the file
    and referred to via the netCDF ``external_variables`` global
    attribute. However, omitted constructs may be written to an
    external file (see the *external* parameter for details).

    **NetCDF unlimited dimensions**

    Domain axis constructs that correspond to NetCDF unlimited
    dimensions may be accessed with the
    `~{{package}}.DomainAxis.nc_is_unlimited` and
    `~{{package}}.DomainAxis.nc_set_unlimited` methods of a domain
    axis construct.

    **NetCDF4 hierarchical groups**

    Hierarchical groups in CF provide a mechanism to structure
    variables within netCDF4 datasets with well defined rules for
    resolving references to out-of-group netCDF variables and
    dimensions. The group structure defined by a field construct's
    netCDF interface will, by default, be recreated in the output
    dataset. See the *group* parameter for details.

    **NetCDF4 dataset chunks**

    Dataset chunking is configured by the *dataset_chunks* parameter,
    which defines the chunking strategy for all output data, including
    the option of no chunking. However, this will be overridden for
    any data that defines its own chunking strategy. See
    `{{package}}.Field.nc_set_hdf5_chunksizes`,
    `{{package}}.Data.nc_set_hdf5_chunksizes`, and `{{package}}.read`.

    .. versionadded:: (cfdm) 1.7.0

    .. seealso:: `{{package}}.read`

    :Parameters:

        fields: (sequence of) `Field` or `Domain`
            The field and domain constructs to write to the file.

        filename: `str`
            The output netCDF file name. Various type of expansion are
            applied to the file names.

            Relative paths are allowed, and standard tilde and shell
            parameter expansions are applied to the string.

            *Parameter example:*
              The file ``file.nc`` in the user's home directory could
              be described by any of the following:
              ``'$HOME/file.nc'``, ``'${HOME}/file.nc'``,
              ``'~/file.nc'``, ``'~/tmp/../file.nc'``.

        fmt: `str`, optional
            The format of the output file. One of:

            ==========================  ==============================
            *fmt*                       Output file type
            ==========================  ==============================
            ``'NETCDF4'``               NetCDF4 format file. This is
                                        the default.

            ``'NETCDF4_CLASSIC'``       NetCDF4 classic format file
                                        (see below)

            ``'NETCDF3_CLASSIC'``       NetCDF3 classic format file
                                        (limited to file sizes less
                                        than 2GB).

            ``'NETCDF3_64BIT_OFFSET'``  NetCDF3 64-bit offset format
                                        file

            ``'NETCDF3_64BIT'``         An alias for
                                        ``'NETCDF3_64BIT_OFFSET'``

            ``'NETCDF3_64BIT_DATA'``    NetCDF3 64-bit offset format
                                        file with extensions (see
                                        below)
            ==========================  ==============================

            By default the format is ``'NETCDF4'``.

            All formats support large files (i.e. those greater than
            2GB) except ``'NETCDF3_CLASSIC'``.

            ``'NETCDF3_64BIT_DATA'`` is a format that requires version
            4.4.0 or newer of the C library (use
            `{{package}}.environment` to see which version of the
            netCDF-C library is in use). It extends the
            ``'NETCDF3_64BIT_OFFSET'`` binary format to allow for
            unsigned 64 bit integer data types and 64-bit dimension
            sizes.

            ``'NETCDF4_CLASSIC'`` files use the version 4 disk format
            (HDF5), but omit features not found in the version 3
            API. They can be read by HDF5 clients. They can also be
            read by netCDF3 clients only if they have been re-linked
            against the netCDF4 library.

            ``'NETCDF4'`` files use the version 4 disk format (HDF5)
            and use the new features of the version 4 API.

        mode: `str`, optional
            Specify the mode of write access for the output file. One of:

            ========  =================================================
            *mode*    Description
            ========  =================================================
            ``'w'``   Open a new file for writing to. If it exists and
                      *overwrite* is True then the file is deleted
                      prior to being recreated.

            ``'a'``   Open an existing file for appending new
                      information to. The new information will be
                      incorporated whilst the original contents of the
                      file will be preserved.

                      In practice this means that new fields will be
                      created, whilst the original fields will not be
                      edited at all. Coordinates on the fields, where
                      equal, will be shared as standard.

                      For append mode, note the following:

                      * Global attributes on the file
                        will remain the same as they were originally,
                        so will become inaccurate where appended fields
                        have incompatible attributes. To rectify this,
                        manually inspect and edit them as appropriate
                        after the append operation using methods such as
                        `nc_clear_global_attributes` and
                        `nc_set_global_attribute`.

                      * Fields with incompatible ``featureType`` to
                        the original file cannot be appended.

                      * At present fields with groups cannot be
                        appended, but this will be possible in a future
                        version. Groups can however be cleared, the
                        fields appended, and groups re-applied, via
                        methods such as `nc_clear_variable_groups` and
                        `nc_set_variable_groups`, to achieve the same
                        for now.

                      * At present domain ancillary constructs of
                        appended fields may not be handled correctly
                        and can appear as extra fields. Set them on the
                        resultant fields using `set_domain_ancillary`
                        and similar methods if required.

            ``'r+'``  Alias for ``'a'``.

            ========  =================================================

            By default the file is opened with write access mode
            ``'w'``.

        overwrite: `bool`, optional
            If False then raise an error if the output file
            pre-exists. By default a pre-existing output file is
            overwritten.

        Conventions: (sequence of) `str`, optional
             Specify conventions to be recorded by the netCDF global
             ``Conventions`` attribute. By default the current
             conventions are always included, but if an older CF
             conventions is defined then this is used instead.

             *Parameter example:*
               ``Conventions='UGRID-1.0'``

             *Parameter example:*
               ``Conventions=['UGRID-1.0']``

             *Parameter example:*
               ``Conventions=['CMIP-6.2', 'UGRID-1.0']``

             *Parameter example:*
               ``Conventions='CF-1.7'``

             *Parameter example:*
               ``Conventions=['CF-1.7', 'UGRID-1.0']``

             Note that if the ``Conventions`` property is set on a
             field construct then it is ignored.

        file_descriptors: `dict`, optional
             Create description of file contents netCDF global
             attributes from the specified attributes and their
             values.

             If any field construct has a property with the same name
             then it will be written as a netCDF data variable
             attribute, even if it has been specified by the
             *global_attributes* parameter, or has been flagged as
             global on any of the field constructs (see
             `{{package}}.Field.nc_global_attributes` for details).

             Identification of the conventions being adhered to by the
             file are not specified as a file descriptor, but by the
             *Conventions* parameter instead.

             *Parameter example:*
               ``file_attributes={'title': 'my data'}``

             *Parameter example:*
               ``file_attributes={'history': 'created 2019-01-01',
               'foo': 'bar'}``

        global_attributes: (sequence of) `str`, optional
             Create netCDF global attributes from the specified field
             construct properties, rather than netCDF data variable
             attributes.

             These attributes are in addition to the following field
             construct properties, which are created as netCDF global
             attributes by default:

             * the description of file contents properties (as defined
               by the CF conventions), and

             * properties flagged as global on any of the field
               constructs being written (see
               `{{package}}.Field.nc_global_attributes` for details).

             Note that it is not possible to create a netCDF global
             attribute from a property that has different values for
             different field constructs being written. In this case
             the property will not be written as a netCDF global
             attribute, even if it has been specified by the
             *global_attributes* parameter or is one of the default
             properties, but will appear as an attribute on the netCDF
             data variable corresponding to each field construct that
             contains the property.

             Any global attributes that are also specified as file
             descriptors will not be written as netCDF global
             variables, but as netCDF data variable attributes
             instead.

             *Parameter example:*
               ``global_attributes='project'``

             *Parameter example:*
               ``global_attributes=['project']``

             *Parameter example:*
               ``global_attributes=['project', 'experiment']``

        variable_attributes: (sequence of) `str`, optional
             Create netCDF data variable attributes from the specified
             field construct properties.

             By default, all field construct properties that are not
             created as netCDF global properties are created as
             attributes netCDF data variables. See the
             *global_attributes* parameter for details.

             Any field construct property named by the
             *variable_attributes* parameter will always be created as
             a netCDF data variable attribute

             *Parameter example:*
               ``variable_attributes='project'``

             *Parameter example:*
               ``variable_attributes=['project']``

             *Parameter example:*
               ``variable_attributes=['project', 'doi']``

        external: `str`, optional
            Write metadata constructs that have data and are marked as
            external to the named external file. Ignored if there are
            no such constructs.

        datatype: `dict`, optional
            Specify data type conversions to be applied prior to
            writing data to disk. This may be useful as a means of
            packing, or because the output format does not support a
            particular data type (for example, netCDF3 classic files
            do not support 64-bit integers). By default, input data
            types are preserved. Any data type conversion is only
            applied to the arrays on disk, and not to the input field
            constructs themselves.

            Data types conversions are defined by `numpy.dtype`
            objects in a dictionary whose keys are input data types
            with values of output data types.

            *Parameter example:*
              To convert 64-bit integers to 32-bit integers:
              ``datatype={numpy.dtype('int64'):
              numpy.dtype('int32')}``.

        endian: `str`, optional
            The endian-ness of the output file. Valid values are
            ``'little'``, ``'big'`` or ``'native'``. By default the
            output is native endian. See the `netCDF4 package
            <http://unidata.github.io/netcdf4-python>`_ for more
            details.

            *Parameter example:*
              ``endian='big'``

        compress: `int`, optional
            Regulate the speed and efficiency of zlib
            compression. Must be an integer between ``0`` and
            ``9``. ``0`` means no compression; ``1`` is the fastest,
            but has the lowest compression ratio; ``9`` is the slowest
            but best compression ratio. The default value is ``4``.

            *Parameter example:*
              ``compress=0``

        least_significant_digit: `int`, optional
            Truncate the input field construct data arrays, but not
            the data arrays of metadata constructs. For a given
            positive integer, N the precision that is retained in the
            compressed data is 10 to the power -N. For example, a
            value of 2 will retain a precision of 0.01. In conjunction
            with the *compress* parameter this produces 'lossy', but
            significantly more efficient, compression. See the
            `netCDF4 package
            <http://unidata.github.io/netcdf4-python>`_ for more
            details.

            *Parameter example:*
              ``least_significant_digit=3``

        fletcher32: `bool`, optional
            If True then the Fletcher-32 HDF5 checksum algorithm is
            activated to detect compression errors. Ignored if
            *compress* is ``0``. See the `netCDF4 package
            <http://unidata.github.io/netcdf4-python>`_ for details.

        shuffle: `bool`, optional
            If True (the default) then the HDF5 shuffle filter (which
            de-interlaces a block of data before compression by
            reordering the bytes by storing the first byte of all of a
            variable's values in the chunk contiguously, followed by
            all the second bytes, and so on) is turned off. By default
            the filter is applied because if the data array values are
            not all wildly different, using the filter can make the
            data more easily compressible.  Ignored if the *compress*
            parameter is ``0`` (which is its default value). See the
            `netCDF4 package
            <http://unidata.github.io/netcdf4-python>`_ for more
            details.

        string: `bool`, optional
            By default string-valued construct data are written as
            netCDF arrays of type string if the output file format is
            ``'NETCDF4'``, or of type char with an extra dimension
            denoting the maximum string length for any other output
            file format (see the *fmt* parameter). If *string* is False
            then string-valued construct data are written as netCDF
            arrays of type char with an extra dimension denoting the
            maximum string length, regardless of the selected output
            file format.

            .. versionadded:: (cfdm) 1.8.0

        verbose: `int` or `str` or `None`, optional
            If an integer from ``-1`` to ``3``, or an equivalent string
            equal ignoring case to one of:

            * ``'DISABLE'`` (``0``)
            * ``'WARNING'`` (``1``)
            * ``'INFO'`` (``2``)
            * ``'DETAIL'`` (``3``)
            * ``'DEBUG'`` (``-1``)

            set for the duration of the method call only as the
            minimum cut-off for the verboseness level of displayed
            output (log) messages, regardless of the
            globally-configured `{{package}}.log_level`.  Note that
            increasing numerical value corresponds to increasing
            verbosity, with the exception of ``-1`` as a special case
            of maximal and extreme verbosity.

            Otherwise, if `None` (the default value), output messages
            will be shown according to the value of the
            `{{package}}.log_level` setting.

            Overall, the higher a non-negative integer or equivalent string
            that is set (up to a maximum of ``3``/``'DETAIL'``) for
            increasing verbosity, the more description that is printed to
            convey how constructs map to output netCDF dimensions, variables
            and attributes.

       warn_valid: `bool`, optional
            If True (the default) then print a warning when writing
            "out-of-range" data, as indicated by the values, if
            present, of any of the ``valid_min``, ``valid_max`` or
            ``valid_range`` properties on field and metadata
            constructs that have data. If False the warning
            is not printed.

            The consequence of writing out-of-range data values is
            that, by default, these values will be masked when the
            file is subsequently read.

            *Parameter example:*
              If a construct has ``valid_max`` property with value
              ``100`` and data with maximum value ``999``, then the
              resulting warning may be suppressed by setting
              ``warn_valid=False``.

            .. versionadded:: (cfdm) 1.8.3

        group: `bool`, optional
            If False then create a "flat" netCDF file, i.e. one with
            only the root group, regardless of any group structure
            specified by the field constructs. By default any groups
            defined by the netCDF interface of the field constructs and
            its components will be created and populated.

            .. versionadded:: (cfdm) 1.8.6

        coordinates: `bool`, optional
            If True then include CF-netCDF coordinate variable names
            in the 'coordinates' attribute of output data
            variables. By default only auxiliary and scalar coordinate
            variables are included.

            .. versionadded:: (cfdm) 1.8.7.0

        omit_data: (sequence of) `str`, optional
            Do not write the data of the named construct types.

            This does not affect the amount of netCDF variables and
            dimensions that are written to the file, nor the netCDF
            variables' attributes, but does not create data on disk
            for the requested variables. The resulting file will be
            smaller than it otherwise would have been, and when the
            new file is read the data of these variables will be
            represented by an array of all missing data.

            The *omit_data* parameter may be one, or a sequence, of:

            ==========================  ===============================
            *omit_data*                 Construct types
            ==========================  ===============================
            ``'field'``                 Field constructs
            ``'field_ancillary'``       Field ancillary constructs
            ``'domain_ancillary'``      Domain ancillary constructs
            ``'dimension_coordinate'``  Dimension coordinate constructs
            ``'auxiliary_coordinate'``  Auxiliary coordinate constructs
            ``'cell_measure'``          Cell measure constructs
            ``'domain_topology'``       Domain topology constructs
            ``'cell_connectivity'``     Cell connectivity constructs
            ``'all'``                   All of the above constructs
            ==========================  ===============================

            *Parameter example:*
              To omit the data from only field constructs:
              ``omit_data='field'`` or ``omit_data=['field']``.

            *Parameter example:*
              To omit the data from domain ancillary and cell measure
              constructs: ``omit_data=['domain_ancillary',
              'cell_measure']``.

            .. versionadded:: (cfdm) 1.10.0.1

        dataset_chunks: `str` or `int` or `float`, optional
            The dataset chunking strategy for data arrays being written
            to the file.

            By default, *dataset_chunks* is ``'4 MiB'``, i.e. 4194304
            bytes.

            If any `Data` being written already stores its own dataset
            chunking strategy (i.e. its `Data.nc_hdf5_chunksizes`
            method returns something other than `None`) then, for that
            data array alone, it is used in preference to the strategy
            defined by the *dataset_chunks* parameter.

            Ignored for netCDF3 output formats, for which all data is
            always written out contiguously.

            .. note:: By default, a data array returned by
                      `{{package}}.read` stores its dataset chunking
                      strategy from the file being read. When this
                      happens that same dataset chunking strategy will
                      be used when the data is written to a new
                      netCDF4 file, unless the strategy was modified
                      or removed prior to writing. To prevent the
                      dataset chunking strategy from the original file
                      being stored, see the *store_dataset_chunks*
                      parameter to `{{package}}.read`.

            The *dataset_chunks* parameter may be one of:

            * ``'contiguous'``

              The data will written to the file contiguously, i.e. no
              chunking.

            * `int` or `float` or `str`

              The size in bytes of the dataset chunks. A floating
              point value is rounded down to the nearest integer, and
              a string represents a quantity of byte
              units. "Square-like" chunk shapes are preferred,
              maximising the amount of chunks that are completely
              filled with data values. For instance a chunksize of
              1024 bytes may be specified with any of ``1024``,
              ``1024.9``, ``'1024'``, ``'1024.9'``, ``'1024 B'``, ``'1
              KiB'``, ``'0.001024 MB'``, etc. Recognised byte units
              are (case insensitive): ``B``, ``KiB``, ``MiB``,
              ``GiB``, ``TiB``, ``PiB``, ``KB``, ``MB``, ``GB``,
              ``TB``, and ``PB``. Spaces in strings are optional.

            .. note:: When the dataset chunk size is defined by a
                      number of bytes (taken either from the
                      *dataset_chunks* parameter, or as stored by the
                      data itself), "square-like" dataset chunk shapes
                      are preferred that maximise the amount of chunks
                      that are completely filled with data values. For
                      example, with *dataset_chunks* of ``'4 MiB'``, a
                      data array of 64-bit floats with shape (400,
                      300, 60) will be written with 20 dataset chunks,
                      each of which contains (93, 93, 60)
                      elements. The first axis is split across 5
                      chunks, the second axis across 4 chunks, and the
                      third axis across 1 chunk containing 60
                      elements. 12 of these chunks are completely
                      filled with 93*93*60 data values (93*93*60*8 B =
                      3.9592 MiB), whilst the remaining 8 chunks at
                      the "edges" of the array contain only 93*21*60,
                      28*93*60, or 28*21*60 data values. The shape of
                      the dataset chunks is based on the shape of the
                      data array and its data type, and is calculated
                      internally with the
                      `dask.array.core.normalize_chunks` function. The
                      use of native compression (see the *compress*
                      parameter) does not affect the dataset chunk
                      size.

            .. versionadded:: (cfdm) 1.12.0.0

        cfa: `str` or `dict` or `None`, optional
            Specify which netCDF variables, if any, should be written
            as CF-netCDF aggregation variables.

            By default, *cfa* is the string ``'auto'``, meaning that a
            construct that was previously read from a CF-netCDF
            aggregation variable will be written as an aggregation
            variable, provided that its data have not been changed in
            ways which prevent the encoding, in which case a normal
            non-aggregation variable will be written. This default, as
            well as other options, can be configured by setting *cfa*
            to a dictionary.

            .. note:: If the intention is to create aggregation
                      variables from fields read from disk, then it is
                      strongly recommended to use the *cfa_write*
                      parameter to `{{package}}.read`, in order to set
                      up the conditions in which a CF-netCDF
                      aggregation encoding is possible.

            If *cfa* is `None` or a (sequence of) `str` then it
            defines which types of constructs are to be written as CF
            aggregation variables:

            ==========================  ===============================
            *cfa*                       Constructs
            ==========================  ===============================
            ``'auto'``                  **This is the default**. Any
                                        construct whose data is
                                        unchanged from having been
                                        previously read from a
                                        CF-netCDF aggregation
                                        variable.

            `None`                      No aggregation variables will
                                        be created.

            ``'field'``                 Field constructs
            ``'field_ancillary'``       Field ancillary constructs
            ``'domain_ancillary'``      Domain ancillary constructs
            ``'dimension_coordinate'``  Dimension coordinate constructs
            ``'auxiliary_coordinate'``  Auxiliary coordinate constructs
            ``'cell_measure'``          Cell measure constructs
            ``'domain_topology'``       Domain topology constructs
            ``'cell_connectivity'``     Cell connectivity constructs
            ``'all'``                   All constructs
            ==========================  ===============================

            If *cfa* is a dictionary then it is used to explicitly
            configure the writing of aggregation variables. It may
            have some or all of the following keys:

            * ``'constructs'``: `None`, `dict` or (sequence of) `str`

              The types of construct to be written as aggregation
              variables.

              If the value is `None` or a (sequence of) `str` then the
              types are the same as if the *cfa* parameter itself was
              set to that value (see the table above). If the
              ``'constructs'`` key is missing then ``'auto'`` is
              assumed.

              If the value is a `dict` then each of its keys defines a
              construct type (see the table above), with a value that
              specifies the number of dimensions that a construct of
              that type must also have if it is to be written as an
              aggregation variable. A value of `None` means no
              restriction on the number of dimensions.

              *Example:*
                Equivalent ways to only write cell measure constructs
                as aggregation variables: ``{'constructs':
                'cell_measure'}``, ``{'constructs': ['cell_measure']}``,
                ``{'cell_measure': None}``.

              *Example:*
                Equivalent ways to only write field and auxiliary
                coordinate constructs as aggregation variables:
                ``{'constructs': ('field', 'auxiliary_coordinate')}``
                and ``{'constructs': {'field': None,
                'auxiliary_coordinate': None}}``.

              *Example:*
                To only write two-dimensional auxiliary coordinate
                constructs as aggregation variables: ``{'constructs':
                {'auxiliary_coordinate': 2}}``.

              *Example:*
                Write two-dimensional auxiliary coordinate constructs
                as aggregation variables, and also all field
                constructs: ``{'constructs': {'auxiliary_coordinate':
                2, 'field': None}}``.

              *Example:*
                 Write any three-dimensional construct whose data is
                 unchanged from having been previously read from a
                 CF-netCDF aggregation variable: ``{'constructs':
                 {'auto': 3}}``.

            * ``'uri'``: `str`

              Specify the URI format of the fragment file names.

              If ``'default'`` (the default) then the fragment file
              names will be written with the same URI formats that
              they had when read from input files (for file names
              originating from the reading of normal non-aggregation
              variables, this will result in absolute URIs). If
              ``'absolute'`` then all fragment file names will be
              written as absolute URIs. If ``'relative'`` then all
              fragment file names will be written as relative-path URI
              references URIs, relative to the location of the
              aggregation file.

            * ``'strict'``: `bool`

              If True (the default if this key is missing) then an
              exception is raised if it is not possible to create an
              aggregation variable from any construct identified by
              the ``'constructs'`` option. If False then a normal,
              non-aggregation variable will be written in this case.

            .. versionadded:: (cfdm) 1.11.2.0

        _implementation: (subclass of) `CFDMImplementation`, optional
            Define the CF data model implementation that defines field
            and metadata constructs and their components.

    :Returns:

        `None`

    **Examples**

    >>> {{package}}.write(f, 'file.nc')

    >>> {{package}}.write(f, 'file.nc', fmt='NETCDF3_CLASSIC')

    >>> {{package}}.write(f, 'file.nc', external='cell_measures.nc')

    >>> {{package}}.write(f, 'file.nc', Conventions='CMIP6')

    """

    implementation = implementation()

    def __new__(
        cls,
        fields,
        filename,
        fmt="NETCDF4",
        mode="w",
        overwrite=True,
        global_attributes=None,
        variable_attributes=None,
        file_descriptors=None,
        external=None,
        Conventions=None,
        datatype=None,
        single=False,
        double=False,
        least_significant_digit=None,
        endian="native",
        compress=4,
        fletcher32=False,
        shuffle=True,
        reference_datetime=None,
        string=True,
        verbose=None,
        warn_valid=True,
        group=True,
        coordinates=False,
        omit_data=None,
        dataset_chunks="4 MiB",
        cfa="auto",
        extra_write_vars=None,
    ):
        """Write field and domain constructs to a netCDF file."""
        # Flatten the sequence of intput fields
        fields = tuple(cls._flat(fields))
        if not fields:
            raise ValueError(
                "Must provide at least one Field or Domain to be written "
            )

        # Parse double and single
        if datatype and (single or double):
            raise ValueError(
                "Can't set 'datatype' at the same time as "
                "'single' or 'double'"
            )

        if single:
            if double:
                raise ValueError(
                    "Can't set both the 'single' and 'double' parameters"
                )

            datatype = {
                np.dtype(float): np.dtype("float32"),
                np.dtype(int): np.dtype("int32"),
            }

        if double:
            datatype = {
                np.dtype("float32"): np.dtype(float),
                np.dtype("int32"): np.dtype(int),
            }

        netcdf = NetCDFWrite(cls.implementation)
        netcdf.write(
            fields,
            filename,
            fmt=fmt,
            mode=mode,
            overwrite=overwrite,
            global_attributes=global_attributes,
            variable_attributes=variable_attributes,
            file_descriptors=file_descriptors,
            external=external,
            Conventions=Conventions,
            datatype=datatype,
            least_significant_digit=least_significant_digit,
            endian=endian,
            compress=compress,
            shuffle=shuffle,
            reference_datetime=reference_datetime,
            fletcher32=fletcher32,
            string=string,
            verbose=verbose,
            warn_valid=warn_valid,
            group=group,
            coordinates=coordinates,
            extra_write_vars=extra_write_vars,
            omit_data=omit_data,
            dataset_chunks=dataset_chunks,
            cfa=cfa,
        )



from . import core, mixin


class List(
    mixin.NetCDFVariable,
    mixin.PropertiesData,
    mixin.Files,
    core.abstract.PropertiesData,
):
    """A list variable required to uncompress a gathered array.

    Compression by gathering combines axes of a multidimensional array
    into a new, discrete axis whilst omitting the missing values and
    thus reducing the number of values that need to be stored.

    The information needed to uncompress the data is stored in a list
    variable that gives the indices of the required points.

    **NetCDF interface**

    {{netCDF variable}}

    .. versionadded:: (cfdm) 1.7.0

    """

    def __init__(
        self,
        properties=None,
        data=None,
        source=None,
        copy=True,
        _use_data=True,
    ):
        """**Initialisation**

        :Parameters:

            {{init properties: `dict`, optional}}

                *Parameter example:*
                  ``properties={'long_name': 'uncompression indices'}``

            {{init data: data_like, optional}}

            {{init source: optional}}

            {{init copy: `bool`, optional}}

        """
        super().__init__(
            properties=properties,
            data=data,
            source=source,
            copy=copy,
            _use_data=_use_data,
        )

        self._initialise_netcdf(source)
        self._initialise_original_filenames(source)

    def dump(
        self,
        display=True,
        _key=None,
        _title=None,
        _create_title=True,
        _prefix="",
        _level=0,
        _omit_properties=None,
        _axes=None,
        _axis_names=None,
    ):
        """A full description of the list variable.

        Returns a description of all properties, including those of
        components, and provides selected values of all data arrays.

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            display: `bool`, optional
                If False then return the description as a string. By
                default the description is printed.

        :Returns:

            {{returns dump}}

        """
        if _create_title and _title is None:
            _title = "List: " + self.identity(default="")

        return super().dump(
            display=display,
            _key=_key,
            _omit_properties=_omit_properties,
            _prefix=_prefix,
            _level=_level,
            _title=_title,
            _create_title=_create_title,
            _axes=_axes,
            _axis_names=_axis_names,
        )



import logging
from copy import deepcopy

import numpy as np

from . import core, mixin
from .data import Data
from .decorators import _manage_log_level_via_verbosity

logger = logging.getLogger(__name__)


class CellMethod(mixin.Container, core.CellMethod):
    """A cell method construct of the CF data model.

    One or more cell method constructs describe how the cell values of
    the field construct represent the variation of the physical
    quantity within its cells, i.e. the structure of the data at a
    higher resolution.

    A single cell method construct consists of a set of axes, a
    "method" property which describes how a value of the field
    construct's data array describes the variation of the quantity
    within a cell over those axes (e.g. a value might represent the
    cell area average), and descriptive qualifiers serving to indicate
    more precisely how the method was applied (e.g. recording the
    spacing of the original data, or the fact that the method was
    applied only over El Nino years).

    .. versionadded:: (cfdm) 1.7.0

    """

    def __new__(cls, *args, **kwargs):
        """Store component classes.

        .. versionadded:: (cfdm) 1.8.7.0

        """
        instance = super().__new__(cls)
        instance._Data = Data
        return instance

    def __str__(self):
        """Called by the `str` built-in function.

        x.__str__() <==> str(x)

        Returns a CF-netCDF-like string of the cell method.

        Note that if the intention is to use this string in a CF-netCDF
        cell_methods attribute then, unless they are standard names, the
        axes names will need to be modified to be netCDF dimension names.

        .. versionadded:: (cfdm) 1.7.0

        """
        string = [f"{axis}:" for axis in self.get_axes(())]

        string.append(self.get_method(""))

        for portion in ("within", "where", "over"):
            q = self.get_qualifier(portion, None)
            if q is not None:
                string.extend((portion, q))

        interval = self.get_qualifier("interval", ())
        comment = self.get_qualifier("comment", None)

        if interval:
            x = ["("]

            y = [f"interval: {data}" for data in interval]
            x.append(" ".join(y))

            if comment is not None:
                x.append(f" comment: {comment}")

            x.append(")")

            string.append("".join(x))

        elif comment is not None:
            string.append(f"({comment})")

        return " ".join(string)

    def _identities_iter(self):
        """Return all possible identities.

        See `identities` for details and examples.

        :Returns:

            generator
                The identities.

        """
        n = self.get_method(None)
        if n is not None:
            yield f"method:{n}"

    def creation_commands(
        self, namespace=None, indent=0, string=True, name="c", header=True
    ):
        """Returns the commands to create the cell method construct.

        .. versionadded:: (cfdm) 1.8.7.0

        .. seealso:: `{{package}}.Data.creation_commands`,
                     `{{package}}.Field.creation_commands`

        :Parameters:

            {{namespace: `str`, optional}}

            {{indent: `int`, optional}}

            {{string: `bool`, optional}}

            {{name: `str`, optional}}

            {{header: `bool`, optional}}

        :Returns:

            {{returns creation_commands}}

        **Examples**

        >>> x = {{package}}.CellMethod(
        ...     axes=['area'],
        ...     qualifiers={'over': 'land'}
        ... )
        >>> print(x.creation_commands(header=False))
        c = {{package}}.CellMethod()
        c.set_axes(('area',))
        c.set_qualifier('over', 'land')

        """
        namespace0 = namespace
        if namespace is None:
            namespace = self._package() + "."
        elif namespace and not namespace.endswith("."):
            namespace += "."

        out = []

        method = self.get_method(None)

        if header:
            out.append("#")
            out.append(f"# {self.construct_type}:")
            if method is not None:
                out[-1] += f" {method}"

        out.append(f"{name} = {namespace}{self.__class__.__name__}()")

        if method is not None:
            out.append(f"{name}.set_method({method!r})")

        axes = self.get_axes(None)
        if axes is not None:
            out.append(f"{name}.set_axes({axes!r})")

        for term, value in self.qualifiers().items():
            if term == "interval":
                value = deepcopy(value)
                for i, data in enumerate(value[:]):
                    if isinstance(data, self._Data):
                        value[i] = data.creation_commands(
                            name=None,
                            namespace=namespace0,
                            indent=0,
                            string=True,
                        )
                    else:
                        value[i] = str(data)

                value = ", ".join(value)
                value = f"[{value}]"
            else:
                value = repr(value)

            out.append(f"{name}.set_qualifier({term!r}, {value})")

        if string:
            indent = " " * indent
            out[0] = indent + out[0]
            out = ("\n" + indent).join(out)

        return out

    def dump(self, display=True, _title=None, _level=0):
        """A full description of the cell method construct.

        Returns a description the method, all qualifiers and the axes
        to which it applies.

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            display: `bool`, optional
                If False then return the description as a string. By
                default the description is printed.

        :Returns:

            {{returns dump}}

        """
        indent0 = "    " * _level

        if _title is None:
            _title = "Cell Method: "

        return indent0 + _title + str(self)

    @_manage_log_level_via_verbosity
    def equals(
        self,
        other,
        rtol=None,
        atol=None,
        verbose=None,
        ignore_qualifiers=(),
        ignore_type=False,
    ):
        """Whether two cell method constructs are the same.

        Equality is strict by default. This means that:

        * the descriptive qualifiers must be the same (see the
          *ignore_qualifiers* parameter).

        The axes of the cell method constructs are *not* considered,
        because they may only be correctly interpreted by the field
        constructs that contain the cell method constructs in
        question. They are, however, taken into account when two
        fields constructs are tested for equality.

        {{equals tolerance}}

        Any type of object may be tested but, in general, equality is
        only possible with another cell method construct, or a
        subclass of one. See the *ignore_type* parameter.

        {{equals tolerance}}

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            other:
                The object to compare for equality.

            {{atol: number, optional}}

            {{rtol: number, optional}}

            {{verbose: `int` or `str` or `None`, optional}}

            ignore_qualifiers: sequence of `str`, optional
                The names of qualifiers to omit from the comparison.

            {{ignore_type: `bool`, optional}}

        :Returns:

            `bool`
                Whether the two cell method constructs are equal.

        **Examples**

        >>> c = {{package}}.CellMethod()
        >>> c.equals(c)
        True
        >>> c.equals(c.copy())
        True
        >>> c.equals('not a cell method')
        False

        """
        pp = super()._equals_preprocess(
            other, verbose=verbose, ignore_type=ignore_type
        )
        if pp is True or pp is False:
            return pp

        other = pp

        # ------------------------------------------------------------
        # Check the methods
        # ------------------------------------------------------------
        if self.get_method(None) != other.get_method(None):
            logger.info(
                f"{self.__class__.__name__}: Different methods: "
                f"{self.get_method(None)!r} != {other.get_method(None)!r}"
            )  # pragma: no cover
            return False

        # ------------------------------------------------------------
        # Check the qualifiers
        # ------------------------------------------------------------
        self_qualifiers = self.qualifiers()
        other_qualifiers = other.qualifiers()

        for prop in tuple(ignore_qualifiers) + ("interval",):
            self_qualifiers.pop(prop, None)
            other_qualifiers.pop(prop, None)

        if set(self_qualifiers) != set(other_qualifiers):
            for q in set(self_qualifiers).symmetric_difference(
                other_qualifiers
            ):
                logger.info(
                    f"{self.__class__.__name__}: Non-common qualifier: {q!r}"
                )  # pragma: no cover
            return False

        for qualifier, x in self_qualifiers.items():
            y = other_qualifiers[qualifier]

            if not self._equals(
                x,
                y,
                rtol=rtol,
                atol=atol,
                ignore_data_type=True,
                verbose=verbose,
                basic=True,
            ):
                logger.info(
                    f"{self.__class__.__name__}: Different {prop} "
                    f"qualifiers: {x!r}, {y!r}"
                )  # pragma: no cover
                return False

        if "interval" in ignore_qualifiers:
            return True

        intervals0 = self.get_qualifier("interval", ())
        intervals1 = other.get_qualifier("interval", ())
        if intervals0:
            if not intervals1:
                logger.info(
                    f"{self.__class__.__name__}: Different interval "
                    f"qualifiers: {intervals0!r} != {intervals1!r}"
                )  # pragma: no cover
                return False

            if len(intervals0) != len(intervals1):
                logger.info(
                    f"{self.__class__.__name__}: Different numbers of "
                    f"interval qualifiers: {intervals0!r} != {intervals1!r}"
                )  # pragma: no cover
                return False

            for data0, data1 in zip(intervals0, intervals1):
                if not self._equals(
                    data0,
                    data1,
                    rtol=rtol,
                    atol=atol,
                    verbose=verbose,
                    ignore_data_type=True,
                    ignore_fill_value=True,
                ):
                    logger.info(
                        f"{self.__class__.__name__}: Different interval "
                        f"qualifiers: {intervals0!r} != {intervals1!r}"
                    )  # pragma: no cover
                    return False

        elif intervals1:
            logger.info(
                f"{self.__class__.__name__}: Different intervals: "
                f"{intervals0!r} != {intervals1!r}"
            )  # pragma: no cover
            return False

        # ------------------------------------------------------------
        # Do NOT check the axes
        # ------------------------------------------------------------

        return True

    def identity(self, default=""):
        """Return the canonical identity for the cell method construct.

        By default the identity is the first found of the following:

        1. The method, preceded by 'method:'
        2. The value of the *default* parameter.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `identities`

        :Parameters:

            default: optional
                If no identity can be found then return the value of the
                default parameter.

        :Returns:

                The identity.

        **Examples**

        >>> f = {{package}}.example_field(1)
        >>> c = f.get_construct('cellmethod1')
        >>> c.get_method()
        'maximum'
        >>> c.identity()
        'method:maximum'
        >>> c.del_method()
        'maximum'
        >>> c.identity()
        ''
        >>> c.identity(default='no identity')
        'no identity'

        """
        n = self.get_method(None)
        if n is not None:
            return f"method:{n}"

        return default

    def identities(self, generator=False, **kwargs):
        """Return all possible identities.

        The identities comprise:

        * The method, preceded by 'method:'.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `identity`

        :Parameters:

            generator: `bool`, optional
                If True then return a generator for the identities,
                rather than a list.

                .. versionadded:: (cfdm) 1.8.9.0

            kwargs: optional
                Additional configuration parameters that may be used
                by subclasses.

                .. versionadded:: (cfdm) 1.8.9.0

        :Returns:

            `list` or generator
                The identities.

        **Examples**

        >>> f = {{package}}.example_field(1)
        >>> c = f.get_construct('cellmethod1')
        >>> c.get_method()
        'maximum'
        >>> c.identities()
        ['method:maximum']
        >>> c.del_method()
        'maximum'
        >>> c.identities()
        []
        >>> for i in c.identities(generator=True):
        ...     print(i)
        ...

        """
        g = self._iter(body=self._identities_iter(), **kwargs)
        if generator:
            return g

        return list(g)

    def sorted(self, indices=None):
        """Return a new cell method construct with sorted axes.

        The axes are sorted by domain axis construct identifier or
        standard name, and any intervals are sorted accordingly.

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            indices: ordered sequence of `int`, optional
                Sort the axes with the given indices. By default the axes
                are sorted by domain axis construct identifier or standard
                name.

        :Returns:

            `CellMethod`
                A new cell method construct with sorted axes.

        **Examples**

        >>> cm = {{package}}.CellMethod(axes=['domainaxis1', 'domainaxis0'],
        ...                      method='mean',
        ...                      qualifiers={'interval': [1, 2]})
        >>> cm
        <{{repr}}CellMethod: domainaxis1: domainaxis0: mean (interval: 1 interval: 2)>
        >>> cm.sorted()
        <{{repr}}CellMethod: domainaxis0: domainaxis1: mean (interval: 2 interval: 1)>

        >>> cm = {{package}}.CellMethod(axes=['domainaxis0', 'area'],
        ...                      method='mean',
        ...                      qualifiers={'interval': [1, 2]})
        >>> cm
        <{{repr}}CellMethod: domainaxis0: area: mean (interval: 1 interval: 2)>
        >>> cm.sorted()
        <{{repr}}CellMethod: area: domainaxis0: mean (interval: 2 interval: 1)>

        """
        new = self.copy()

        axes = new.get_axes(())
        if len(axes) == 1:
            return new

        if indices is None:
            indices = np.argsort(axes)
        elif len(indices) != len(axes):
            raise ValueError(
                f"Can't sort cell method axes. The given indices ({indices}) "
                f"do not correspond to the number of axes ({axes})"
            )

        axes2 = []
        for i in indices:
            axes2.append(axes[i])

        new.set_axes(tuple(axes2))

        intervals = new.get_qualifier("interval", ())
        if len(intervals) <= 1:
            return new

        intervals2 = []
        for i in indices:
            intervals2.append(intervals[i])

        new.set_qualifier("interval", tuple(intervals2))

        return new



from .abstract import SubsampledSubarray


class InterpolationSubarray(SubsampledSubarray):
    """A subarray of an array compressed by subsamplng.

    A subarray describes a unique part of the uncompressed array.

    The subarray has no standardised interpolation algorithm. An
    interpolation algorithm may have been provided via the
    *interpolation_description* parameter, but that algorithm can not
    be interpreted by the {{class}} object. Nonetheless, the subarray
    is still representable in its uncompressed form (e.g. its shape is
    known), even though the uncompressed data values can not be
    calculated.

    See CF section 8.3.3 "Interpolation Variable".

    .. versionadded:: (cfdm) 1.10.0.2

    """

    def __getitem__(self, indices):
        """Return a subspace of the uncompressed data.

        x.__getitem__(indices) <==> x[indices]

        Always raises a `ValueError`, rather than returning a subspace
        of the uncompressed data as an independent numpy array, since
        the non-standardised interpolation algorithm can't be
        interpreted.

        .. versionadded:: (cfdm) 1.10.0.2

        """
        raise ValueError(
            "Can't uncompress subsampled data using a non-standardised "
            "interpolation algorithm:\n\n"
            f"{self.get_interpolation_description('')}"
        )



from .linearinterpolation import LinearInterpolation


class BiLinearInterpolation(LinearInterpolation):
    """Mixin class for bi-linear interpolation.

    See CF appendix J "Coordinate Interpolation Methods".

    .. versionadded:: (cfdm) 1.10.0.0

    """

    def _bilinear_interpolation(self, ua, ub, uc, ud, d2, d1):
        """Interpolate bi-linearly between four points.

        uac = fl(ua, uc, s(ia2, ic2, i2))
        ubd = fl(ub, ud, s(ia2, ic2, i2))
        u(i2, i1) = fl(uac, ubd, s(ia1, ib1, i1))

        See CF appendix J "Coordinate Interpolation Methods".

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `_linear_interpolation`, `_s`

        :Parameters:

            ua, ub: `numpy.ndarray`
                The tie points at locations A and B, in the sense of
                CF appendix J Figure J.2.

            uc, ud: `numpy.ndarray`
                The tie points at locations C and D, in the sense of
                CF appendix J Figure J.2.

            {{d2: `int`}}

            {{d1: `int`}}

        :Returns:

            `numpy.ndarray`
                The result of interpolating the tie points to interior
                locations implied by *s*.

        """
        uac = self._linear_interpolation(ua, uc, d2)
        ubd = self._linear_interpolation(ub, ud, d2)

        u = self._linear_interpolation(uac, ubd, d1)

        return u



import numpy as np

from . import abstract


class DimensionCoordinate(abstract.Coordinate):
    """A dimension coordinate construct of the CF data model.

    A dimension coordinate construct provides information which locate
    the cells of the domain and which depend on a subset of the domain
    axis constructs. The dimension coordinate construct is able to
    unambiguously describe cell locations because a domain axis can be
    associated with at most one dimension coordinate construct, whose
    data array values must all be non-missing and strictly
    monotonically increasing or decreasing. They must also all be of
    the same numeric data type. If cell bounds are provided, then each
    cell must have exactly two vertices. CF-netCDF coordinate
    variables and numeric scalar coordinate variables correspond to
    dimension coordinate constructs.

    The dimension coordinate construct consists of a data array of the
    coordinate values which spans a subset of the domain axis
    constructs, an optional array of cell bounds recording the extents
    of each cell (stored in a `Bounds` object), and properties to
    describe the coordinates. An array of cell bounds spans the same
    domain axes as its coordinate array, with the addition of an extra
    dimension whose size is that of the number of vertices of each
    cell. This extra dimension does not correspond to a domain axis
    construct since it does not relate to an independent axis of the
    domain. Note that, for climatological time axes, the bounds are
    interpreted in a special way indicated by the cell method
    constructs.

    .. versionadded:: (cfdm) 1.7.0

    """

    @property
    def construct_type(self):
        """Return a description of the construct type.

        .. versionadded:: (cfdm) 1.7.0

        :Returns:

            `str`
                The construct type.

        **Examples**

        >>> c = {{package}}.{{class}}()
        >>> c.construct_type
        'dimension_coordinate'

        """
        return "dimension_coordinate"

    def set_data(self, data, copy=True, inplace=True):
        """Set the data.

        The units, calendar and fill value of the incoming `Data` instance
        are removed prior to insertion.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `data`, `del_data`, `get_data`, `has_data`

        :Parameters:

            data: data_like
                The data to be inserted. Must be 1-dimensional,
                i.e. scalar or multidimensional data is not allowed.

                {{data_like}}

            {{data copy: `bool`, optional}}

            {{inplace: `bool`, optional (default True)}}

                .. versionadded:: (cfdm) 1.8.7.0

        :Returns:

            `None` or `{{class}}`
                If the operation was in-place then `None` is returned,
                otherwise return a new `{{class}}` instance containing the
                new data.

        **Examples**

        >>> d = {{package}}.Data(range(10))
        >>> f.set_data(d)
        >>> f.has_data()
        True
        >>> f.get_data()
        <{{repr}}Data(10): [0, ..., 9]>
        >>> f.del_data()
        <{{repr}}Data(10): [0, ..., 9]>
        >>> f.has_data()
        False
        >>> print(f.get_data(None))
        None
        >>> print(f.del_data(None))
        None

        """
        if np.ndim(data) != 1:
            raise ValueError(
                "Dimension coordinate construct must have 1-dimensional data. "
                f"Got {data!r}"
            )

        return super().set_data(data, copy=copy, inplace=inplace)



"""Flatten NetCDF groups.

Portions of this package were adapted from the `netcdf_flattener`
library, which carries the following Apache 2.0 License:

Copyright (c) 2020 EUMETSAT

Licensed to the Apache Software Foundation (ASF) under one or more
contributor license agreements. The ASF licenses this file to you
under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy
of the License at http://www.apache.org/licenses/LICENSE-2.0.

"""

from .flatten import netcdf_flatten



from os.path import join

from uritools import urisplit

from ...functions import abspath
from ..abstract import FileArray
from ..mixin import IndexMixin
from .mixin import FragmentArrayMixin


class FragmentFileArray(
    FragmentArrayMixin,
    IndexMixin,
    FileArray,
):
    """Fragment of aggregated data in a file.

    .. versionadded:: (cfdm) 1.12.0.0

    """

    def __new__(cls, *args, **kwargs):
        """Store fragment classes.

        .. versionadded:: (cfdm) 1.12.0.0

        """
        # Import fragment classes. Do this here (as opposed to outside
        # the class) to aid subclassing.
        from . import FragmentH5netcdfArray, FragmentNetCDF4Array

        instance = super().__new__(cls)
        instance._FragmentArrays = (
            FragmentNetCDF4Array,
            FragmentH5netcdfArray,
        )
        return instance

    def __init__(
        self,
        filename=None,
        address=None,
        dtype=None,
        shape=None,
        storage_options=None,
        unpack_aggregated_data=True,
        aggregated_attributes=None,
        aggregation_file_directory=None,
        source=None,
        copy=True,
    ):
        """**Initialisation**

        :Parameters:

            filename: (sequence of `str`), optional
                The locations of fragment datasets containing the
                array.

            address: (sequence of `str`), optional
                How to find the array in the fragment datasets.

            dtype: `numpy.dtype`, optional
                The data type of the aggregated array. May be `None`
                if is not known. This may differ from the data type of
                the fragment's data.

            shape: `tuple`, optional
                The shape of the fragment in its canonical form.

            {{init attributes: `dict` or `None`, optional}}

                If *attributes* is `None`, the default, then the
                attributes will be set from the fragment dataset
                during the first `__getitem__` call.

            {{aggregated_units: `str` or `None`, optional}}

            {{aggregated_calendar: `str` or `None`, optional}}

            {{init storage_options: `dict` or `None`, optional}}

            {{init source: optional}}

            {{init copy: `bool`, optional}}

        """
        super().__init__(
            filename=filename,
            address=address,
            dtype=dtype,
            shape=shape,
            mask=True,
            unpack=True,
            attributes=None,
            storage_options=storage_options,
            source=source,
            copy=copy,
        )

        if source is not None:
            try:
                aggregated_attributes = source._get_component(
                    "aggregated_attributes", None
                )
            except AttributeError:
                aggregated_attributes = None

            try:
                unpack_aggregated_data = source._get_component(
                    "unpack_aggregated_data", True
                )
            except AttributeError:
                unpack_aggregated_data = True

            try:
                aggregation_file_directory = source._get_component(
                    "aggregation_file_directory", None
                )
            except AttributeError:
                aggregation_file_directory = None

        self._set_component(
            "aggregation_file_directory",
            aggregation_file_directory,
            copy=False,
        )
        self._set_component(
            "unpack_aggregated_data",
            unpack_aggregated_data,
            copy=False,
        )
        if aggregated_attributes is not None:
            self._set_component(
                "aggregated_attributes", aggregated_attributes, copy=copy
            )

    def _get_array(self, index=None):
        """Returns a subspace of the dataset variable.

        The method acts as a factory for either a
        `NetCDF4FragmentArray`, `H5netcdfFragmentArray`, or
        `UMFragmentArray` class, and it is the result of calling
        `!_get_array` on the newly created instance that is returned.

        `H5netcdfFragmentArray` will only be used if
        `NetCDF4FragmentArray` returns a `FileNotFoundError`
        exception; and `UMFragmentArray` will only be used
        if `H5netcdfFragmentArray` returns an `Exception`.

        .. versionadded:: (cfdm) 1.12.0.0

        .. seealso:: `__array__`, `index`

        :Parameters:

            {{index: `tuple` or `None`, optional}}

               When a `tuple`, there must be a distinct entry for each
               fragment dimension.

        :Returns:

            `numpy.ndarray`
                The subspace.

        """
        # Loop round the fragment array backends, in the order
        # given by the `_FragmentArrays` attribute (which is
        # defined in `__new__`), until we find one that can open
        # the file.
        if index is None:
            index = self.index()

        errors = []
        for FragmentArray in self._FragmentArrays:
            try:
                array = FragmentArray(source=self, copy=False)._get_array(
                    index
                )
            except Exception as error:  # noqa: F841
                errors.append(
                    f"{FragmentArray().__class__.__name__}:\n"
                    f"{error.__class__.__name__}: {error}"
                )
            else:
                return array

        # Still here?
        error = "\n\n".join(errors)
        raise RuntimeError(
            f"Can't access array index {index} from fragment file "
            f"{self.get_filename()}:\n\n"
            f"{error}"
        )

    def get_filename(self, normalise=False, default=AttributeError()):
        """The name of the file containing the fragment.

        .. versionadded:: (cfdm) 1.12.0.0

        :Parameters:

            {{normalise: `bool`, optional}}

            default: optional
                Return the value of the *default* parameter if there
                is no file name.

                {{default Exception}}

        :Returns:

            `str`
                The file name.

        """
        filename = super().get_filename(normalise=False, default=None)
        if filename is None:
            if default is None:
                return

            return self._default(
                default, f"{self.__class__.__name__} has no file name"
            )

        if normalise:
            uri = urisplit(filename)

            # Convert the file name to an absolute URI
            if uri.isrelpath():
                # File name is a relative-path URI reference
                filename = join(
                    self._get_component("aggregation_file_directory"), filename
                )
            elif not (uri.isabsuri() or uri.isabspath()):
                raise ValueError(
                    "Fragment file location must be an absolute URI, a "
                    "relative-path URI reference, or an absolute-path URI: "
                    f"Got: {filename!r}"
                )

            filename = abspath(filename, uri=True)

        return filename



import abc


class Implementation(metaclass=abc.ABCMeta):
    """Store an implementation of the CF data model."""

    def __init__(self, cf_version=None, **kwargs):
        """**Initialisation**

        :Parameters:

            kwargs:
                The concrete objects required to represent a Field.

        """
        self._cf_version = cf_version
        self._class = kwargs.copy()
        for key, value in kwargs.items():
            if value is None:
                del self._class[key]

    def classes(self):
        """Return all the classes of the implementation.

        :Returns:

            `dict`
                The class objects, keyed by their class name.

        **Examples**

        >>> i = cfdm.Implementation()  # an abstract implementation
        >>> i.classes()
        {}

        >>> i = cfdm.implementation()  # child CF data model implementation
        >>> sorted(i.classes())
        ['AuxiliaryCoordinate',
         'Bounds',
         'CellMeasure',
         'CellMethod',
         'CoordinateConversion',
         'CoordinateReference',
         'Count',
         'Data',
         'Datum',
         'DimensionCoordinate',
         'Domain',
         'DomainAncillary',
         'DomainAxis',
         'Field',
         'FieldAncillary',
         'GatheredArray',
         'Index',
         'InteriorRing',
         'List',
         'NetCDF4Array',
         'NodeCountProperties',
         'PartNodeCountProperties',
         'RaggedContiguousArray',
         'RaggedIndexedArray',
         'RaggedIndexedContiguousArray']

        """
        return self._class.copy()

    def copy(self):
        """Copy."""
        return type(self)(cf_version=self.get_cf_version(), **self._class)

    def get_class(self, name):
        """Return a class of the implementation.

        :Parameters:

            name: `str`
                The name of the class.

                *Parameter example:*
                  ``name='Field'``

        :Returns:

                The class object.

        **Examples**

        >>> i = cfdm.implementation()  # child CF data model implementation
        >>> Field = i.get_class('Field')
        >>> f = Field()

        """
        try:
            return self._class[name]
        except KeyError:
            raise ValueError(f"Implementation does not have class {name!r}")

    def get_cf_version(self):
        """Return the CF version of the implementation.

        :Returns:

            `str`
                The version.

        **Examples**

        >>> i = cfdm.implementation()  # child CF data model implementation
        >>> i.get_cf_version()
        '1.10'

        """
        return self._cf_version

    def set_class(self, name, cls):
        """Set a class of the implementation.

        :Parameters:

            name: `str`
                The name of the class.

                *Parameter example:*
                  ``name='Field'``

            cls:
                The class object.

        :Returns:

            `None`

        **Examples**

        >>> i = cfdm.Implementation()  # child CF data model implementation
        >>> i.classes()
        {}

        >>> i.set_class('Field', cfdm.Field)
        >>> i.classes()
        {'Field': <class 'cfdm.field.Field'>}

        >>> field_class = i.get_class('Field')
        >>> f = field_class()
        >>> f
        <Field: >

        """
        self._class[name] = cls



"""cfdm is a reference implementation of the CF data model.

It identifies the fundamental elements of the CF conventions and
shows how they relate to each other, independently of the netCDF
encoding.

The central element defined by the CF data model is the field
construct, which corresponds to a CF-netCDF data variable with all of
its metadata.

The cfdm package implements the CF data model for its internal data
structures and so is able to process any CF-compliant dataset. It is
not strict about CF-compliance, however, so that partially conformant
datasets may be modified in memory, as well as ingested from existing
datasets and written to new datasets. This is so that datasets which
are partially conformant may nonetheless be modified in memory and
written to new datasets.

The cfdm package can:

    * read field constructs from netCDF and CDL datasets,
    * create new field constructs in memory,
    * write and append field constructs to netCDF datasets on disk,
    * read, write, and create datasets containing hierarchical groups,
    * read, write, and create coordinates defined by geometry cells,
    * inspect field constructs,
    * test whether two field constructs are the same,
    * modify field construct metadata and data,
    * create subspaces of field constructs,
    * incorporate, and create, metadata stored in external files, and
    * read, write, and create data that have been compressed by
      convention (i.e. ragged or gathered arrays), whilst presenting a
      view of the data in its uncompressed form.

Note that cfdm enables the creation of CF field constructs, but it's
up to the user to use them in a CF-compliant way.

"""

import logging
import sys

from packaging.version import Version

from . import core

__date__ = core.__date__
__cf_version__ = core.__cf_version__
__version__ = core.__version__

_requires = core._requires + (
    "cftime",
    "netCDF4",
    "dask",
    "scipy",
    "h5netcdf",
    "s3fs",
    "uritools",
    "cfunits",
)

_error0 = f"cfdm requires the modules {', '.join(_requires)}. "

# Check the version of cftime
try:
    import cftime
except ImportError as error1:
    raise ImportError(_error0 + str(error1))
else:
    _minimum_vn = "1.6.4"
    if Version(cftime.__version__) < Version(_minimum_vn):
        raise ValueError(
            f"Bad cftime version: cfdm requires cftime>={_minimum_vn}. "
            f"Got {cftime.__version__} at {cftime.__file__}"
        )

# Check the version of netCDF4
try:
    import netCDF4
except ImportError as error1:
    raise ImportError(_error0 + str(error1))
else:
    _minimum_vn = "1.7.2"
    if Version(netCDF4.__version__) < Version(_minimum_vn):
        raise ValueError(
            f"Bad netCDF4 version: cfdm requires netCDF4>={_minimum_vn}. "
            f"Got {netCDF4.__version__} at {netCDF4.__file__}"
        )

# Check the version of h5netcdf
try:
    import h5netcdf
except ImportError as error1:
    raise ImportError(_error0 + str(error1))
else:
    _minimum_vn = "1.3.0"
    if Version(h5netcdf.__version__) < Version(_minimum_vn):
        raise ValueError(
            f"Bad h5netcdf version: cfdm requires h5netcdf>={_minimum_vn}. "
            f"Got {h5netcdf.__version__} at {h5netcdf.__file__}"
        )

# Check the version of h5py
try:
    import h5py
except ImportError as error1:
    raise ImportError(_error0 + str(error1))
else:
    _minimum_vn = "3.12.0"
    if Version(h5py.__version__) < Version(_minimum_vn):
        raise ValueError(
            f"Bad h5py version: cfdm requires h5py>={_minimum_vn}. "
            f"Got {h5py.__version__} at {h5py.__file__}"
        )

# Check the version of s3fs
try:
    import s3fs
except ImportError as error1:
    raise ImportError(_error0 + str(error1))
else:
    _minimum_vn = "2024.6.0"
    if Version(s3fs.__version__) < Version(_minimum_vn):
        raise ValueError(
            f"Bad s3fs version: cfdm requires s3fs>={_minimum_vn}. "
            f"Got {s3fs.__version__} at {s3fs.__file__}"
        )

# Check the version of scipy
try:
    import scipy
except ImportError as error1:
    raise ImportError(_error0 + str(error1))
else:
    _minimum_vn = "1.10.0"
    if Version(scipy.__version__) < Version(_minimum_vn):
        raise ValueError(
            f"Bad scipy version: cfdm requires scipy>={_minimum_vn}. "
            f"Got {scipy.__version__} at {scipy.__file__}"
        )

# Check the version of dask
try:
    import dask
except ImportError as error1:
    raise ImportError(_error0 + str(error1))
else:
    _minimum_vn = "2025.2.0"
    _maximum_vn = "2025.3.0"
    if Version(dask.__version__) < Version(_minimum_vn) or Version(
        dask.__version__
    ) > Version(_maximum_vn):
        raise ValueError(
            "Bad dask version: cfdm requires "
            f"dask>={_minimum_vn},<={_maximum_vn}. "
            f"Got {dask.__version__} at {dask.__file__}"
        )

# Check the version of uritools
try:
    import uritools
except ImportError as error1:
    raise ImportError(_error0 + str(error1))
else:
    _minimum_vn = "4.0.3"
    if Version(uritools.__version__) < Version(_minimum_vn):
        raise ValueError(
            f"Bad uritools version: cfdm requires uritools>={_minimum_vn}. "
            f"Got {uritools.__version__} at {uritools.__file__}"
        )

# Check the version of cfunits
try:
    import cfunits
except ImportError as error1:
    raise ImportError(_error0 + str(error1))
else:
    _minimum_vn = "3.3.7"
    if Version(cfunits.__version__) < Version(_minimum_vn):
        raise ValueError(
            f"Bad cfunits version: cfdm requires cfunits>={_minimum_vn}. "
            f"Got {cfunits.__version__} at {cfunits.__file__}"
        )

del _minimum_vn

from .constants import masked

# Internal ones passed on so they can be used in cf-python (see
# comment below)
from .functions import (
    ATOL,
    CF,
    LOG_LEVEL,
    RTOL,
    abspath,
    atol,
    chunksize,
    configuration,
    dirname,
    environment,
    integer_dtype,
    log_level,
    parse_indices,
    rtol,
    unique_constructs,
    _disable_logging,
    _reset_log_emergence_level,
    _is_valid_log_level_int,
    Configuration,
    Constant,
    ConstantAccess,
    is_log_level_debug,
    is_log_level_detail,
    is_log_level_info,
)

# Though these are internal-use methods, include them in the namespace
# (without documenting them) so that cf-python can use them internally
# too:
from .decorators import (
    _inplace_enabled,
    _inplace_enabled_define_and_cleanup,
    _manage_log_level_via_verbosity,
    _display_or_return,
)

from .constructs import Constructs

from .data import (
    Array,
    AggregatedArray,
    BoundsFromNodesArray,
    CellConnectivityArray,
    CompressedArray,
    Data,
    FullArray,
    GatheredArray,
    H5netcdfArray,
    NetCDFArray,
    NetCDF4Array,
    netcdf_indexer,
    NumpyArray,
    PointTopologyArray,
    RaggedArray,
    RaggedContiguousArray,
    RaggedIndexedArray,
    RaggedIndexedContiguousArray,
    SparseArray,
    SubsampledArray,
)

from .data import (
    BiLinearSubarray,
    BiQuadraticLatitudeLongitudeSubarray,
    BoundsFromNodesSubarray,
    CellConnectivitySubarray,
    GatheredSubarray,
    InterpolationSubarray,
    LinearSubarray,
    QuadraticLatitudeLongitudeSubarray,
    QuadraticSubarray,
    RaggedSubarray,
    Subarray,
    SubsampledSubarray,
)

from .count import Count
from .index import Index
from .interpolationparameter import InterpolationParameter
from .list import List
from .nodecountproperties import NodeCountProperties
from .partnodecountproperties import PartNodeCountProperties
from .tiepointindex import TiePointIndex

from .bounds import Bounds
from .coordinateconversion import CoordinateConversion
from .datum import Datum
from .interiorring import InteriorRing

from .units import Units

from .auxiliarycoordinate import AuxiliaryCoordinate
from .cellconnectivity import CellConnectivity
from .cellmeasure import CellMeasure
from .cellmethod import CellMethod
from .coordinatereference import CoordinateReference
from .dimensioncoordinate import DimensionCoordinate
from .domain import Domain
from .domainancillary import DomainAncillary
from .domainaxis import DomainAxis
from .domaintopology import DomainTopology
from .field import Field
from .fieldancillary import FieldAncillary

from .abstract import Implementation
from .cfdmimplementation import CFDMImplementation, implementation

from .read_write import read, write
from .read_write.netcdf.flatten import netcdf_flatten

from .examplefield import example_field, example_fields, example_domain

from .abstract import Container

# --------------------------------------------------------------------
# Set up basic logging for the full project with a root logger
# --------------------------------------------------------------------
# Configure the root logger which all module loggers inherit from:
logging.basicConfig(
    stream=sys.stdout,
    style="{",  # default is old style ('%') string formatting
    format="{message}",  # no module names or datetimes etc. for basic case
    level=logging.WARNING,  # default but change level via log_level()
)

# And create custom level inbetween 'INFO' & 'DEBUG', to understand
# value see:
# https://docs.python.org/3.8/howto/logging.html#logging-levels
logging.DETAIL = 15  # set value as an attribute as done for built-in levels
logging.addLevelName(logging.DETAIL, "DETAIL")


def detail(self, message, *args, **kwargs):
    """Sets up a custom logging level named 'detail'."""
    if self.isEnabledFor(logging.DETAIL):
        self._log(logging.DETAIL, message, args, **kwargs)


logging.Logger.detail = detail



from copy import deepcopy

import numpy as np
from cfunits import Units


class ArrayMixin:
    """Mixin class for a container of an array.

    .. versionadded:: (cfdm) 1.11.2.0

    """

    # Functions handled by __array_function__ implementations (numpy
    # NEP 18)
    _HANDLED_FUNCTIONS = {}

    def __array__(self, dtype=None, copy=None):
        """The numpy array interface.

        .. versionadded:: (cfdm) 1.8.7.0

        :Parameters:

            dtype: optional
                Typecode or data-type to which the array is cast.

            copy: `None` or `bool`
                Included to match the v2 `numpy.ndarray.__array__`
                API, but ignored. The return numpy array is always
                independent.

                .. versionadded:: (cfdm) 1.12.0.0

        :Returns:

            `numpy.ndarray`
                An independent numpy array of the data.

        **Examples**

        >>> isinstance(a, Array)
        True
        >>> n = numpy.asanyarray(a)
        >>> isinstance(n, numpy.ndarray)
        True

        """
        array = self.array
        if dtype is None:
            return array

        return array.astype(dtype, copy=False)

    def __array_function__(self, func, types, args, kwargs):
        """Implement the `numpy` ``__array_function__`` protocol.

        See numpy NEP 18 for details.

        .. versionadded:: (cfdm) 1.11.2.0

        """
        if func not in self._HANDLED_FUNCTIONS:
            return NotImplemented

        # Note: This allows subclasses that don't override
        #       __array_function__ to handle Array objects
        if not all(issubclass(t, self.__class__) for t in types):
            return NotImplemented

        return self._HANDLED_FUNCTIONS[func](*args, **kwargs)

    def __getitem__(self, indices):
        """Return a subspace of the uncompressed subarray.

        x.__getitem__(indices) <==> x[indices]

        Returns a subspace of the uncompressed subarray as an
        independent numpy array.

        .. versionadded:: (cfdm) 1.8.7.0

        """
        raise NotImplementedError(
            f"Must implement {self.__class__.__name__}.__getitem__"
        )  # pragma: no cover

    def __repr__(self):
        """Called by the `repr` built-in function.

        x.__repr__() <==> repr(x)

        .. versionadded:: (cfdm) 1.8.7.0

        """
        return f"<{self.__class__.__name__}{self.shape}: {self}>"

    def __str__(self):
        """Called by the `str` built-in function.

        x.__str__() <==> str(x)

        .. versionadded:: (cfdm) 1.8.7.0

        """
        return f"shape={self.shape}, dtype={self.dtype}"

    def __docstring_package_depth__(self):
        """Returns the package depth for {{package}} substitutions.

        See `_docstring_package_depth` for details.

        """
        return 0

    @property
    def _meta(self):
        """Normalize the array to an appropriate Dask meta object.

        The Dask meta can be thought of as a suggestion to Dask. Dask
        uses this meta to generate the task graph until it can infer
        the actual metadata from the values. It does not force the
        output to have the structure or dtype of the specified meta.

        .. versionadded:: (cfdm) 1.11.2.0

        .. seealso:: `dask.utils.meta_from_array`

        """
        return np.array((), dtype=self.dtype)

    @property
    def Units(self):
        """The `Units` object containing the units of the array.

        .. versionadded:: (cfdm) 1.11.2.0

        """
        return Units(self.get_units(None), self.get_calendar(None))

    def astype(self, dtype, **kwargs):
        """Cast the data to a specified type.

        .. versionadded:: (cfdm) 1.12.1.0

        :Parameters:

            dtype: `str` or dtype
                Typecode or data-type to which the array is cast.

            kwargs: optional
                Any other keywords accepted by `np.astype`.

        :Returns:

            `np.ndarray`
                The data with the new data type

        """
        kwargs["copy"] = False
        return self.array.astype(dtype, **kwargs)

    def get_attributes(self, copy=True):
        """The attributes of the array.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            default: optional
                Return the value of the *default* parameter if the
                attributes have not been set. If set to an `Exception`
                instance then it will be raised instead.

        :Returns:

            `dict`
                The attributes.

        """
        attributes = self._get_component("attributes", None)
        if attributes is None:
            attributes = {}
        elif copy:
            attributes = deepcopy(attributes)

        return attributes

    def get_calendar(self, default=ValueError()):
        """The calendar of the array.

        If the calendar is `None` then the CF default calendar is
        assumed, if applicable.

        .. versionadded:: (cfdm) 1.10.0.1

        :Parameters:

            default: optional
                Return the value of the *default* parameter if the
                calendar has not been set. If set to an `Exception`
                instance then it will be raised instead.

        :Returns:

            `str` or `None`
                The calendar value.

        """
        attributes = self.get_attributes({})
        if "calendar" not in attributes:
            if default is None:
                return

            return self._default(
                default,
                f"{self.__class__.__name__} 'calendar' has not been set",
            )

        return attributes["calendar"]

    def get_compression_type(self):
        """Returns the array's compression type.

        Specifically, returns the type of compression that has been
        applied to the underlying array.

        .. versionadded:: (cfdm) 1.8.7.0

        :Returns:

            `str`
                The compression type. An empty string means that no
                compression has been applied.

        **Examples**

        >>> a.compression_type
        ''

        >>> a.compression_type
        'gathered'

        >>> a.compression_type
        'ragged contiguous'

        """
        return self._get_component("compression_type", "")

    def get_units(self, default=ValueError()):
        """The units of the array.

        If the units are `None` then the array has no defined units.

        .. versionadded:: (cfdm) 1.10.0.1

        .. seealso:: `get_calendar`

        :Parameters:

            default: optional
                Return the value of the *default* parameter if the
                units have not been set. If set to an `Exception`
                instance then it will be raised instead.

        :Returns:

            `str` or `None`
                The units value.

        """
        attributes = self.get_attributes({})
        if "units" not in attributes:
            if default is None:
                return

            return self._default(
                default,
                f"{self.__class__.__name__} 'units' have not been set",
            )

        return attributes["units"]


# --------------------------------------------------------------------
# __array_function__ implementations (numpy NEP 18)
# --------------------------------------------------------------------
def array_implements(cls, numpy_function):
    """Decorator for __array_function__ implementations.

    .. versionadded:: (cfdm) 1.12.0.0

    """

    def decorator(func):
        cls._HANDLED_FUNCTIONS[numpy_function] = func
        return func

    return decorator


@array_implements(ArrayMixin, np.concatenate)
def concatenate(arrays, axis=0):
    """Version of `np.concatenate` that works for `Array` objects.

    .. versionadded:: (cfdm) 1.12.0.0

    """
    # Convert to numpy arrays, and concatenate those.
    return np.ma.concatenate(tuple(map(np.asanyarray, arrays)), axis=axis)



from . import core, mixin


class Count(
    mixin.NetCDFVariable,
    mixin.NetCDFDimension,
    mixin.NetCDFSampleDimension,
    mixin.PropertiesData,
    mixin.Files,
    core.abstract.PropertiesData,
):
    """A count variable required to uncompress a ragged array.

    A collection of features stored using a contiguous ragged array
    combines all features along a single dimension (the sample
    dimension) such that each feature in the collection occupies a
    contiguous block.

    The information needed to uncompress the data is stored in a count
    variable that gives the size of each block.

    **NetCDF interface**

    {{netCDF variable}}

    The name of the netCDF dimension spanned by the count variable's
    data may be accessed with the `nc_set_dimension`,
    `nc_get_dimension`, `nc_del_dimension` and `nc_has_dimension`
    methods.

    The name of the netCDF sample dimension spanned by the compressed
    data (that is stored in the "sample_dimension" netCDF attribute
    and which does not correspond to a domain axis construct) may be
    accessed with the `nc_set_sample_dimension`,
    `nc_get_sample_dimension`, `nc_del_sample_dimension` and
    `nc_has_sample_dimension` methods.

    {{netCDF HDF5 chunks}}

    .. versionadded:: (cfdm) 1.7.0

    """

    def __init__(
        self,
        properties=None,
        data=None,
        source=None,
        copy=True,
        _use_data=True,
    ):
        """**Initialisation**

        :Parameters:

            {{init properties: `dict`, optional}}

                *Parameter example:*
                  ``properties={'long_name': 'number of obs'}``

            {{init data: data_like, optional}}

                {{data_like}}

            {{init source: optional}}

            {{init copy: `bool`, optional}}

        """
        super().__init__(
            properties=properties,
            data=data,
            source=source,
            copy=copy,
            _use_data=_use_data,
        )

        self._initialise_netcdf(source)
        self._initialise_original_filenames(source)

    def dump(
        self,
        display=True,
        _key=None,
        _title=None,
        _create_title=True,
        _prefix="",
        _level=0,
        _omit_properties=None,
        _axes=None,
        _axis_names=None,
    ):
        """A full description of the count variable.

        Returns a description of all properties, including those of
        components, and provides selected values of all data arrays.

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            display: `bool`, optional
                If False then return the description as a string. By
                default the description is printed.

        :Returns:

            {{returns dump}}

        """
        if _create_title and _title is None:
            _title = "Count: " + self.identity(default="")

        return super().dump(
            display=display,
            _key=_key,
            _omit_properties=_omit_properties,
            _prefix=_prefix,
            _level=_level,
            _title=_title,
            _create_title=_create_title,
            _axes=_axes,
            _axis_names=_axis_names,
        )



from .abstract import MeshArray
from .mixin import CompressedArrayMixin
from .subarray import BoundsFromNodesSubarray


class BoundsFromNodesArray(CompressedArrayMixin, MeshArray):
    """An array of cell bounds defined by UGRID node coordinates.

    The UGRID node coordinates contain the locations of the nodes of
    the domain topology. In UGRID, the bounds of edge or face cells
    may be defined by the these locations in conjunction with a
    mapping from from each cell boundary vertex to its corresponding
    coordinate value.

    .. versionadded:: (cfdm) 1.11.0.0

    .. seealso:: `BoundsFromNodesSubarray`

    """

    def __new__(cls, *args, **kwargs):
        """Store subarray classes.

        .. versionadded:: (cfdm) 1.11.0.0

        """
        instance = super().__new__(cls)
        instance._Subarray = {"bounds from nodes": BoundsFromNodesSubarray}
        return instance

    def __init__(
        self,
        node_connectivity=None,
        shape=None,
        start_index=None,
        cell_dimension=None,
        node_coordinates=None,
        source=None,
        copy=True,
    ):
        """**Initialisation**

        :Parameters:

            node_connectivity: array_like
                A 2-d integer array that contains indices which map
                each cell boundary vertex to its corresponding
                position in the 1-d *node_coordinates* array, as found
                in a UGRID "edge_node_connectivity" or
                "face_node_connectivity" variable.

            shape: `tuple`
                The shape of the bounds array.

            {{init start_index: `int`}}

            {{init cell_dimension: `int`}}

            node_coordinates: array_like
                A 1-d array that contains a coordinate location for
                each mesh node, as found in a UGRID "node_coordinates"
                variable.

            {{init source: optional}}

            {{init copy: `bool`, optional}}

        """
        if shape is None and node_connectivity is not None:
            shape = node_connectivity.shape

        # Note: Setting compressed_dimensions={1: (1,)} means that
        #       only one subarray will ever span the trailing
        #       dimension of the 'node_connectivity' array, but the
        #       leading dimension may be chunked.
        super().__init__(
            connectivity=node_connectivity,
            shape=shape,
            start_index=start_index,
            cell_dimension=cell_dimension,
            compressed_dimensions={1: (1,)},
            compression_type="bounds from nodes",
            source=source,
            copy=copy,
        )

        if source is not None:
            try:
                node_coordinates = source.get_node_coordinates(None)
            except AttributeError:
                node_coordinates = None

        if node_coordinates is not None:
            self._set_component(
                "node_coordinates", node_coordinates, copy=copy
            )

    @property
    def dtype(self):
        """The data-type of the uncompressed data.

        .. versionadded:: (cfdm) 1.11.0.0

        """
        return self.get_node_coordinates().dtype

    def conformed_data(self):
        """The conformed node connectivity and node coordinate data.

        The conformed data arrays are mutually consistent and are
        suitable of use in `Subarray` classes.

        .. versionadded:: (cfdm) 1.11.0.0

        :Returns:

            `dict`
                The node connectivity data, with the key ``'data'``,
                and the node coordinate data with the key
                ``'node_coordinates'``.

        """
        out = super().conformed_data()
        out["node_coordinates"] = self.get_node_coordinates()
        return out

    def get_node_coordinates(self, default=ValueError()):
        """The coordinates representing the node locations.

        .. versionadded:: (cfdm) 1.11.0.0

        :Parameters:

            default: optional
                Return the value of the *default* parameter if node
                coordinates indices have not been set.

                {{default Exception}}

        :Returns:

                The node coordinates.

        """
        return self._get_component("node_coordinates", default=default)

    def to_memory(self):
        """Bring data on disk into memory.

        There is no change to data that is already in memory.

        .. versionadded:: (cfdm) 1.11.0.0

        :Returns:

            `{{class}}`
                A copy of the array with all of its data in memory.

        """
        a = super().to_memory()

        node_coordinates = self.get_node_coordinates(None)
        if node_coordinates is not None:
            try:
                a._set_component(
                    "node_coordinates",
                    node_coordinates.to_memory(),
                    copy=False,
                )
            except AttributeError:
                pass

        return a



import copy
import datetime
import faulthandler
import re
import unittest

faulthandler.enable()  # to debug seg faults and timeouts

import cfdm


class ConstructsTest(unittest.TestCase):
    """Unit test for the Constructs class."""

    f = cfdm.example_field(1)
    c = f.constructs

    def setUp(self):
        """Preparations called immediately before each test method."""
        # Disable log messages to silence expected warnings
        cfdm.LOG_LEVEL("DISABLE")
        # Note: to enable all messages for given methods, lines or
        # calls (those without a 'verbose' option to do the same)
        # e.g. to debug them, wrap them (for methods, start-to-end
        # internally) as follows:
        #
        # cfdm.LOG_LEVEL('DEBUG')
        # < ... test code ... >
        # cfdm.log_level('DISABLE')

    def test_Constructs__repr__str__dump(self):
        """Test all means of Constructs inspection."""
        c = self.c

        repr(c)
        str(c)

        c = cfdm.Constructs()
        str(c)

    def test_Constructs__len__(self):
        """Test the length returned for Constructs."""
        c = self.c

        self.assertEqual(len(c), 20)
        self.assertEqual(len(self.f.domain.constructs), 17)

    def test_Constructs_check_construct_type(self):
        """Test the `check_construct_type` Constructs method."""
        c = self.c

        with self.assertRaises(ValueError):
            c._check_construct_type("bad type")

    def test_Constructs_construct_type(self):
        """Test the `construct_type` Constructs method."""
        d = self.f.domain.constructs
        self.assertIsNone(d.construct_type("cell_method"))

    def test_Constructs_construct_types(self):
        """Test the `construct_types` Constructs method."""
        d = self.f.domain.constructs
        self.assertEqual(len(d.construct_types()), len(d))

    def test_Constructs_items_key_value(self):
        """Test the items, key and value Constructs methods."""
        c = self.c

        for i, (key, value) in enumerate(c.items()):
            x = c.filter_by_key(key)
            self.assertEqual(x.key(), key)
            self.assertIs(x.value(), value)

        self.assertEqual(i, 19)

    def test_Constructs_copy_shallow_copy(self):
        """Test the Constructs methods for copying."""
        c = self.c

        d = c.filter_by_type("domain_axis")
        e = d.copy()

        self.assertTrue(d.equals(e, verbose=3))
        self.assertTrue(e.equals(d, verbose=3))

        d = c.shallow_copy()
        self.assertTrue(c.equals(d, verbose=3))
        self.assertTrue(d.equals(c, verbose=3))

    def test_Constructs_domain_axis_identity(self):
        """Test the `domain_axis_identity` Constructs method."""
        c = self.c

        with self.assertRaises(ValueError):
            c.domain_axis_identity(999)

    def test_Constructs_filter(self):
        """Trivial test of constructs-filtering Constructs methods."""
        c = self.c

        for todict in (False, True):
            d = c.filter(
                filter_by_axis=(),
                filter_by_data=True,
                filter_by_identity=(),
                filter_by_key=(),
                filter_by_measure=(),
                filter_by_method=(),
                filter_by_naxes=(),
                filter_by_ncdim=(),
                filter_by_ncvar=(),
                filter_by_property={},
                filter_by_size=(),
                filter_by_type=(),
                todict=todict,
            )
            self.assertEqual(len(d), 0)

        self.assertEqual(c.filter(cached=999), 999)
        self.assertTrue(c.filter().equals(c))

        with self.assertRaises(TypeError):
            c.filter(bad_kwarg=None)

    def test_Constructs_FILTERING(self):
        """Test Constructs methods that invert or reverse filters."""
        c = self.c

        # Axis
        for axis_mode in ("and", "or", "exact", "subset"):
            for args in (
                ["qwerty"],
                ["domainaxis0"],
                ["domainaxis0", "domainaxis1"],
                ["domainaxis0", "domainaxis1", "domainaxis2"],
            ):
                d = c.filter_by_axis(*args, axis_mode=axis_mode)
                e = d.inverse_filter()
                self.assertEqual(len(e), len(c) - len(d))

        # Inverse filter, filters applied
        self.assertEqual(len(c.filters_applied()), 0)
        ci = c.inverse_filter()
        self.assertEqual(len(ci), 0)
        self.assertEqual(len(ci), len(c) - len(c))

        d = c.filter_by_type("dimension_coordinate", "auxiliary_coordinate")
        self.assertEqual(len(d.filters_applied()), 1)
        di = d.inverse_filter()
        self.assertEqual(len(di), len(c) - len(d))

        e = d.filter_by_property(units="degrees")
        self.assertEqual(len(e.filters_applied()), 2)
        ei = e.inverse_filter(1)
        self.assertEqual(len(e.filters_applied()), 2)
        self.assertEqual(len(ei), len(d) - len(e))

        d2 = c.filter_by_type("auxiliary_coordinate")
        e2 = d2.filter_by_naxes(1)
        f2 = e2.inverse_filter(1)
        g2 = f2.inverse_filter(1)
        h2 = g2.inverse_filter(1)
        self.assertTrue(g2.equals(e2, verbose=3))
        self.assertTrue(h2.equals(f2, verbose=3))

        # Unfilter
        self.assertTrue(e.unfilter(1).equals(d, verbose=3))
        self.assertTrue(e.unfilter(1).unfilter().equals(c, verbose=3))
        self.assertTrue(d.unfilter(1).equals(c, verbose=3))
        self.assertTrue(c.unfilter(1).equals(c, verbose=3))

    def test_Constructs_domain_axes(self):
        """Test the `domain_axes` Constructs method."""
        c = self.c

        self.assertEqual(c.domain_axes(cached=999), 999)
        self.assertEqual(
            len(c.domain_axes(filter_by_identity=("grid_longitude",))), 1
        )

        keys = tuple(c.domain_axes(todict=True))
        self.assertEqual(len(c.domain_axes(*keys)), 4)
        self.assertEqual(len(c.domain_axes(keys)), 0)

        with self.assertRaises(TypeError):
            c.domain_axes(filter_by_type=("dimension_coordinate",))

        with self.assertRaises(TypeError):
            c.domain_axes(
                "grid_latitude", filter_by_identity=("grid_longitude",)
            )

    def test_Constructs_filter_by_data(self):
        """Test the `filter_by_data` Constructs method."""
        c = self.c

        self.assertEqual(len(c.filter_by_data()), 12)
        self.assertEqual(c.filter_by_data(cached=999), 999)

    def test_Constructs_filter_by_type(self):
        """Test the `filter_by_type` Constructs method."""
        c = self.c

        self.assertEqual(c.filter_by_type(cached=999), 999)

        d = c.filter(
            filter_by_property={"standard_name": None},
            filter_by_type=("dimension_coordinate",),
            todict=True,
        )
        self.assertEqual(len(d), 4)

        constructs = c.filter_by_type(
            "auxiliary_coordinate",
            "cell_measure",
            "cell_method",
            "coordinate_reference",
            "dimension_coordinate",
            "domain_ancillary",
            "domain_axis",
            "field_ancillary",
        )
        n = 20
        self.assertEqual(len(constructs), n)

        constructs = c.filter_by_type("auxiliary_coordinate")
        n = 3
        self.assertEqual(len(constructs), n)
        for key, value in constructs.items():
            self.assertIsInstance(value, cfdm.AuxiliaryCoordinate)

        constructs = c.filter_by_type("cell_measure")
        n = 1
        self.assertEqual(len(constructs), n)
        for key, value in constructs.items():
            self.assertIsInstance(value, cfdm.CellMeasure)

        constructs = c.filter_by_type("cell_method")
        n = 2
        self.assertEqual(len(constructs), n)
        for key, value in constructs.items():
            self.assertIsInstance(value, cfdm.CellMethod)

        constructs = c.filter_by_type("dimension_coordinate")
        n = 4
        self.assertEqual(len(constructs), n)
        for key, value in constructs.items():
            self.assertIsInstance(value, cfdm.DimensionCoordinate)

        constructs = c.filter_by_type("coordinate_reference")
        n = 2
        self.assertEqual(len(constructs), n)
        for key, value in constructs.items():
            self.assertIsInstance(value, cfdm.CoordinateReference)

        constructs = c.filter_by_type("domain_ancillary")
        n = 3
        self.assertEqual(len(constructs), n)
        for key, value in constructs.items():
            self.assertIsInstance(value, cfdm.DomainAncillary)

        constructs = c.filter_by_type("field_ancillary")
        n = 1
        self.assertEqual(len(constructs), n)
        for key, value in constructs.items():
            self.assertIsInstance(value, cfdm.FieldAncillary)

        constructs = c.filter_by_type("domain_axis")
        n = 4
        self.assertEqual(len(constructs), n)
        for key, value in constructs.items():
            self.assertIsInstance(value, cfdm.DomainAxis)

        constructs = c.filter_by_type(*["domain_ancillary"])
        n = 3
        self.assertEqual(len(constructs), n)
        for key, value in constructs.items():
            self.assertIsInstance(value, cfdm.DomainAncillary)

        constructs = c.filter_by_type("domain_ancillary", "domain_axis")
        n = 7
        self.assertEqual(len(constructs), n)

        self.assertEqual(len(c.filter_by_type("qwerty")), 0)

    def test_Constructs_filter_by_method(self):
        """Test the `filter_by_method` Constructs method."""
        c = self.c

        self.assertEqual(len(c.filter_by_method()), 2)
        self.assertEqual(len(c.filter_by_method("mean")), 1)
        self.assertEqual(len(c.filter_by_method("qwerty")), 0)
        self.assertEqual(c.filter_by_method(cached=999), 999)

    def test_Constructs_filter_by_ncdim(self):
        """Test the `filter_by_ncdim` Constructs method."""
        c = self.c

        self.assertEqual(len(c.filter_by_ncdim()), 4)
        self.assertEqual(len(c.filter_by_ncdim("y")), 1)
        self.assertEqual(len(c.filter_by_ncdim("qwerty")), 0)
        self.assertEqual(c.filter_by_ncdim(cached=999), 999)

    def test_Constructs_filter_by_ncvar(self):
        """Test the `filter_by_ncvar` Constructs method."""
        c = self.c

        self.assertEqual(len(c.filter_by_ncvar()), 14)
        self.assertEqual(len(c.filter_by_ncvar("qwerty")), 0)
        self.assertEqual(len(c.filter_by_ncvar("a")), 1)
        self.assertEqual(c.filter_by_ncvar(cached=999), 999)

    def test_Constructs_filter_by_measure(self):
        """Test the `filter_by_measure` Constructs method."""
        c = self.c

        self.assertEqual(len(c.filter_by_measure()), 1)
        self.assertEqual(len(c.filter_by_measure("qwerty")), 0)
        self.assertEqual(len(c.filter_by_measure("area")), 1)
        self.assertEqual(c.filter_by_measure(cached=999), 999)

    def test_Constructs_filter_by_identity(self):
        """Test the `filter_by_identity` Constructs method."""
        c = self.c

        self.assertEqual(len(c.filter_by_identity()), 20)
        self.assertEqual(len(c.filter_by_identity("qwerty")), 0)
        self.assertEqual(len(c.filter_by_identity("latitude")), 1)
        self.assertEqual(c.filter_by_identity(cached=999), 999)

        self.assertEqual(
            len(c.filter_by_identity("dimensioncoordinate1", "longitude")), 2
        )

        with self.assertRaises(TypeError):
            c("latitude", filter_by_identity=("longitude",))

    def test_Constructs_filter_by_axis(self):
        """Test the `filter_by_axis` Constructs method."""
        c = self.c

        self.assertEqual(len(c.filter_by_axis()), 12)
        self.assertEqual(len(c.filter_by_axis(0, 1, 2, axis_mode="or")), 11)
        self.assertEqual(
            len(c.filter_by_axis("domainaxis1", axis_mode="or")), 7
        )
        self.assertEqual(len(c.filter_by_axis("grid_longitude")), 6)
        self.assertEqual(len(c.filter_by_axis(re.compile("^grid_lon"))), 6)
        self.assertEqual(len(c.filter_by_axis(re.compile("^grid"))), 0)
        self.assertEqual(len(c.filter_by_axis("ncdim%x")), 6)

        self.assertEqual(c.filter_by_axis(cached=999), 999)

        self.assertEqual(len(c.filter_by_axis(99)), 0)
        self.assertEqual(len(c.filter_by_axis(99, todict=True)), 0)

        with self.assertRaises(ValueError):
            c.filter_by_axis(0, 1, axis_mode="bad_mode")

    def test_Constructs_clear_filters_applied(self):
        """Test the `clear_filters_applied` Constructs method."""
        c = self.c

        d = c.shallow_copy()
        d.clear_filters_applied()

    def test_Constructs_filter_by_naxes(self):
        """Test the `filter_by_naxes` Constructs method."""
        c = self.c

        self.assertEqual(len(c.filter_by_naxes()), 12)
        self.assertEqual(len(c.filter_by_naxes(1)), 7)
        self.assertEqual(c.filter_by_naxes(cached=999), 999)

    def test_Constructs_filter_by_property(self):
        """Test the `filter_by_property` Constructs method."""
        c = self.c

        self.assertEqual(len(c.filter_by_property()), 12)

        for mode in ([], ["and"], ["or"]):
            for kwargs in (
                {"qwerty": 34},
                {"standard_name": "surface_altitude"},
                {"standard_name": "surface_altitude", "units": "m"},
                {"standard_name": "surface_altitude", "units": "degrees"},
                {"standard_name": "surface_altitude", "units": "qwerty"},
            ):
                d = c.filter_by_property(*mode, **kwargs)
                e = d.inverse_filter()
                self.assertEqual(len(e), len(c) - len(d))

        self.assertEqual(len(c.filter_by_property(standard_name=None)), 9)

        with self.assertRaises(ValueError):
            c.filter_by_property("too many", "modes")

        with self.assertRaises(ValueError):
            c.filter_by_property("bad_mode")

    def test_Constructs_filter_by_size(self):
        """Test the `filter_by_size` Constructs method."""
        c = self.c

        self.assertEqual(len(c.filter_by_size(9)), 1)
        self.assertEqual(len(c.filter_by_size(9, 10)), 2)
        self.assertEqual(len(c.filter_by_size()), 4)
        self.assertEqual(len(c.filter_by_size(-1)), 0)
        self.assertEqual(c.filter_by_size(cached=999), 999)

    def test_Constructs_filter_by_key(self):
        """Test the `filter_by_key` Constructs method."""
        c = self.c

        self.assertEqual(len(c.filter_by_key()), 20)
        self.assertEqual(len(c.filter_by_key("qwerty")), 0)
        self.assertEqual(len(c.filter_by_key("dimensioncoordinate1")), 1)
        self.assertEqual(len(c.filter_by_key(re.compile("^dim"))), 4)
        self.assertEqual(c.filter_by_key(cached=999), 999)

    def test_Constructs_copy(self):
        """Test the copy module copying behaviour of Constructs."""
        c = self.c

        copy.copy(c)
        copy.deepcopy(c)

    def test_Constructs__getitem__(self):
        """Test the lookup access of construct items from Constructs."""
        c = self.c

        self.assertIsInstance(
            c["auxiliarycoordinate1"], cfdm.AuxiliaryCoordinate
        )

        with self.assertRaises(KeyError):
            c["bad_key"]

    def test_Constructs_get_data_axes(self):
        """Test the `get_data_axes` Constructs method."""
        c = self.c

        with self.assertRaises(ValueError):
            c.get_data_axes("bad key")

    def test_Constructs_todict(self):
        """Test the todict Constructs method."""
        c = self.c

        self.assertIsInstance(c.todict(), dict)

    def test_Constructs_private(self):
        """Test internal (designated as private) Constructs methods."""
        c = self.f.domain.constructs

        # _construct_type_description
        self.assertEqual(
            c._construct_type_description("auxiliary_coordinate"),
            "auxiliary coordinate",
        )

        # _check_construct_type
        self.assertIsNone(c._check_construct_type(None))
        self.assertIsNone(c._check_construct_type("cell_method", None))

        x = c.shallow_copy()
        del x._constructs["auxiliary_coordinate"]
        with self.assertRaises(KeyError):
            x["auxiliarycoordinate1"]

        # _del_construct
        x = self.c.shallow_copy()
        x._del_construct("domainancillary0")

        with self.assertRaises(ValueError):
            x._del_construct("domainaxis1")

        x._del_construct("dimensioncoordinate3")
        with self.assertRaises(ValueError):
            x._del_construct("domainaxis3")

        x = c.shallow_copy()
        x._del_construct("dimensioncoordinate3")
        self.assertIsInstance(x._del_construct("domainaxis3"), cfdm.DomainAxis)

        # _set_construct
        x = self.c.shallow_copy()
        with self.assertRaises(ValueError):
            x._set_construct(
                self.f.construct("cellmethod0"), axes=["domainaxis0"]
            )

        # _set_construct_data_axes
        x = self.c.shallow_copy()
        with self.assertRaises(ValueError):
            x._set_construct_data_axes("qwerty", ["domainaxis"])

        with self.assertRaises(ValueError):
            x._set_construct_data_axes("auxiliarycoordinate1", ["qwerty"])

        with self.assertRaises(ValueError):
            x._set_construct_data_axes("auxiliarycoordinate1", ["domainaxis0"])

        # _pop
        x = c.shallow_copy()
        with self.assertRaises(KeyError):
            x._pop("qwerty")


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cfdm.environment()
    print("")
    unittest.main(verbosity=2)



from . import core, mixin


class CoordinateConversion(
    mixin.ParametersDomainAncillaries, core.CoordinateConversion
):
    """A coordinate conversion component.

    Specifically, a coordinate conversion component of a coordinate
    reference construct of the CF data model.

    A coordinate conversion formula converting coordinate values taken
    from the dimension or auxiliary coordinate constructs to a
    different coordinate system. A term of the conversion formula can
    be a scalar or vector parameter which does not depend on any
    domain axis constructs, may have units, or may be a descriptive
    string (such as the projection name "mercator"), or it can be a
    reference to a domain ancillary construct (such as one containing
    spatially varying orography data). A coordinate reference
    construct relates the coordinate values of the field to locations
    in a planetary reference frame.

    .. versionadded:: (cfdm) 1.7.0

    """



import functools


class cached_property(functools.cached_property):
    """Read only version of functools.cached_property."""

    def __set__(self, instance, val):
        """Raise an error when attempting to set a cached property."""
        raise AttributeError("Can't set attribute")



from ..decorators import _inplace_enabled, _inplace_enabled_define_and_cleanup


class Topology:
    """Mixin class for topology-related constructs.

    .. versionadded:: (cfdm) 1.11.0.0

    """

    @classmethod
    def _normalise_cell_ids(cls, data, start_index, remove_empty_columns):
        """Normalise cell identifier values.

        Normalised data is in a form that is suitable for creating a
        CF-netCDF UGRID connectivity variable.

        The first column of the input data defines the unique
        identifier for each cell.

        See `normalise` for further details.

        .. versionadded:: (cfdm) 1.11.0.0

        .. seealso:: `normalise`

        :Parameters:

            data: `numpy.ndarray`
                The original numpy array of cell identifiers.

            start_index: `int`
                The start index for the data values in the normalised
                data. Must be ``0`` or ``1`` for zero- or one-based
                indices respectively.

            remove_empty_columns: `bool`
                If True then remove any array columns that are
                entirely missing data.

        :Returns:

            `numpy.ndarray`
                The normalised data.

        **Examples*

        See `normalise` for examples.

        """
        import numpy as np

        masked = np.ma.is_masked(data)
        if masked:
            mask = data.mask

        # Get the original cell ids
        ids = data[:, 0]
        id0 = ids[0]
        n_cells = ids.size
        relabel = not (
            (not id0 and (ids == np.arange(n_cells)).all())
            or (id0 == 1 and (ids == np.arange(1, n_cells + 1)).all())
        )

        if not relabel:
            # The cell ids are already correct (barring a possible
            # 'start_index' offset).
            if start_index and not id0:
                data += 1
            elif not start_index and id0 == 1:
                data -= 1

            smallest_id = data[0, 0]
            largest_id = data[-1, 0]
        else:
            # Remove negative values
            #
            # PERFORMANCE WARNING (TODO): A potentially slow loop
            smallest_id = None
            dmin = data.min()
            if dmin < 0:
                data -= dmin
                ids = data[:, 0]

            # Replace the non-negative cell ids (i) with negative
            # numbers (j)
            copyto = np.copyto
            for i, j in zip(ids.tolist(), range(-n_cells, 0)):
                copyto(data, j, where=data == i)

            if masked:
                # Re-mask the data, since np.copyto is not mask-aware.
                data = np.ma.array(data, mask=mask)

            largest_id = -1

        # Remove redundant cell ids. These may occur when a previous
        # indexing operation removed cells.
        move_missing_values = False
        if data.max() > largest_id:
            data = np.ma.where(data > largest_id, np.ma.masked, data)
            move_missing_values = True

        if smallest_id is not None and data.min() < smallest_id:
            data = np.ma.where(data < smallest_id, np.ma.masked, data)
            move_missing_values = True

        if move_missing_values:
            # Move missing values to the end of each row
            data[:, 1:].sort(axis=1, endwith=True)

        if remove_empty_columns:
            # Discard columns that are all missing data
            data = cls._remove_empty_columns(data)

        if relabel:
            # Convert the negative ids to non-negative numbers
            if start_index:
                data += n_cells + 1
            else:
                data += n_cells

        return data

    @classmethod
    def _remove_empty_columns(cls, data):
        """Remove any data columns that are entirely missing data.

        It is assumed that any such columns will be at the end of the
        dimension.

        .. versionadded:: (cfdm) 1.11.0.0

        .. seealso:: `_normalise_cell_ids`, `normalise`

        :Parameters:

            data: `numpy.ndarray`
                The data.

        :Returns:

            `numpy.ndarray`
                The normalised data.

        """
        import numpy as np

        if np.ma.is_masked(data):
            count = data.count(axis=0)
            if not count.min():
                index = np.where(count)[0]
                data = data[:, index[0] : index[-1] + 1]

        return data

    @_inplace_enabled(default=False)
    def transpose(self, axes=None, inplace=False):
        """Permute the axes of the data array.

        In this context, the data only has one axis, the first
        one. The second data dimension can not be moved and may not be
        specified. Therefore, the `transpose` method never changes the
        data.

        .. versionadded:: (cfdm) 1.11.0.0

        .. seealso:: `insert_dimension`, `squeeze`

        :Parameters:

            axes: (sequence of) `int`, optional
                The new axis order. By default the order is reversed.

                {{axes int examples}}

            {{inplace: `bool`, optional}}

        :Returns:

            `{{class}}` or `None`
                The new construct with permuted data axes. If the
                operation was in-place then `None` is returned.

        """
        if axes is None:
            iaxes = list(range(self.ndim - 1, -1, -1))
        else:
            iaxes = self._parse_axes(axes)

        if iaxes != [0]:
            raise ValueError(
                f"Can't transpose {self.__class__.__name__} with axes "
                f"of {axes!r}. Axes must be equivalent to [0]"
            )

        c = _inplace_enabled_define_and_cleanup(self)
        super(Topology, c).transpose(iaxes + [-1], inplace=True)
        return c



import atexit
import datetime
import faulthandler
import os
import tempfile
import unittest

import numpy as np

faulthandler.enable()  # to debug seg faults and timeouts

import cfdm

n_tmpfiles = 1
tmpfiles = [
    tempfile.mkstemp("_test_Domain.nc", dir=os.getcwd())[1]
    for i in range(n_tmpfiles)
]
(tmpfile,) = tmpfiles


def _remove_tmpfiles():
    """Remove temporary files created during tests."""
    for f in tmpfiles:
        try:
            os.remove(f)
        except OSError:
            pass


atexit.register(_remove_tmpfiles)


class DomainTest(unittest.TestCase):
    """Unit test for the Domain class."""

    f = cfdm.example_field(1)
    d = f.domain

    def setUp(self):
        """Preparations called immediately before each test method."""
        # Disable log messages to silence expected warnings
        cfdm.LOG_LEVEL("DISABLE")
        # Note: to enable all messages for given methods, lines or
        # calls (those without a 'verbose' option to do the same)
        # e.g. to debug them, wrap them (for methods, start-to-end
        # internally) as follows:
        #
        # cfdm.LOG_LEVEL('DEBUG')
        # < ... test code ... >
        # cfdm.log_level('DISABLE')

    def test_Domain__repr__str__dump(self):
        """Test all means of Domain inspection."""
        d = self.d

        repr(d)
        str(d)
        self.assertIsInstance(d.dump(display=False), str)

        d.nc_set_variable("domain1")
        for title in (None, "title"):
            d.dump(display=False, _title=title)

        d.nc_del_variable()
        for title in (None, "title"):
            d.dump(display=False, _title=title)

        # Test when any construct which can have data in fact has no data.
        d = d.copy()
        for identity in [
            "time",  # a dimension coordinate
            "latitude",  # an auxiliary coordinate
            "measure:area",  # a cell measure
            "surface_altitude",  # a domain ancillary
        ]:
            c = d.construct(identity)  # get relevant construct, type as above
            c.del_data()
            self.assertFalse(c.has_data())
            str(d)
            repr(d)

    def test_Domain__init__(self):
        """Test the Domain constructor and source keyword."""
        cfdm.Domain(source="qwerty")

    def test_Domain_equals(self):
        """Test the equality-testing Domain method."""
        d = self.d
        e = d.copy()

        self.assertTrue(d.equals(d, verbose=3))
        self.assertTrue(d.equals(e, verbose=3))
        self.assertTrue(e.equals(d, verbose=3))

        e = cfdm.example_field(0).domain
        self.assertFalse(e.equals(d))

        e.set_property("foo", "bar")
        self.assertFalse(e.equals(d))

    def test_Domain_properties(self):
        """Test the property access methods of Domain."""
        d = cfdm.Domain()

        d.set_property("long_name", "qwerty")

        self.assertEqual(d.properties(), {"long_name": "qwerty"})
        self.assertEqual(d.get_property("long_name"), "qwerty")
        self.assertEqual(d.del_property("long_name"), "qwerty")
        self.assertIsNone(d.get_property("long_name", None))
        self.assertIsNone(d.del_property("long_name", None))

        d.set_property("long_name", "qwerty")
        self.assertEqual(d.clear_properties(), {"long_name": "qwerty"})

        d.set_properties({"long_name": "qwerty"})
        d.set_properties({"foo": "bar"})
        self.assertEqual(d.properties(), {"long_name": "qwerty", "foo": "bar"})

    def test_Domain_del_construct(self):
        """Test the `del_construct` Domain method."""
        d = self.d.copy()

        self.assertIsInstance(
            d.del_construct("dimensioncoordinate1"), cfdm.DimensionCoordinate
        )

    def test_Domain_climatological_time_axes(self):
        """Test the `climatological_time_axes` method of Domain."""
        f = cfdm.example_field(7)
        d = f.domain

        self.assertEqual(d.climatological_time_axes(), set())

        cm = cfdm.CellMethod(axes="domainaxis0", method="mean")
        cm.set_qualifier("over", "years")
        f.set_construct(cm)

        self.assertEqual(d.climatological_time_axes(), set(("domainaxis0",)))

    def test_Domain_creation_commands(self):
        """Test the `creation_commands` Domain method."""
        d = self.d.copy()

        with self.assertRaises(ValueError):
            d.creation_commands(name="c")

        with self.assertRaises(ValueError):
            d.creation_commands(name="data", data_name="data")

        d.nc_set_global_attribute("foo", "bar")
        d = d.creation_commands(namespace="my_cfdm", header=True)

    def test_Domain_identity(self):
        """Test the `identity` Domain method."""
        d = self.d

        d.nc_set_variable("qwerty")
        self.assertEqual(d.identity(), "ncvar%qwerty")

        d.set_property("long_name", "qwerty")
        self.assertEqual(d.identity(), "long_name=qwerty")

    def test_Domain_identites(self):
        """Test the `identities` Domain method."""
        d = self.d.copy()

        d.nc_set_variable("qwerty")
        d.set_property("cf_role", "qwerty")
        d.set_property("long_name", "qwerty")
        d.set_property("foo", "bar")

        self.assertEqual(
            d.identities(),
            ["cf_role=qwerty", "long_name=qwerty", "foo=bar", "ncvar%qwerty"],
        )

    def test_Domain_apply_masking(self):
        """Test the `apply_masking` Domain method."""
        d = self.d.copy()

        self.assertIsNone(d.apply_masking(inplace=True))
        self.assertTrue(d.equals(d.apply_masking()))

    def test_Domain_data(self):
        """Test that Domain instances do not have data."""
        self.assertFalse(self.d.has_data())

    def test_Domain_has_bounds(self):
        """Test that Domain instances do not have cell bounds."""
        self.assertFalse(self.d.has_bounds())

    def test_Domain_get_data_axes(self):
        """Test Domain.get_data_axes."""
        d = cfdm.example_domain(0)
        self.assertEqual(d.get_data_axes("latitude"), ("domainaxis0",))

        with self.assertRaises(ValueError):
            d.get_data_axes(None)

        with self.assertRaises(ValueError):
            d.get_data_axes("domainaxis0")

    def test_Domain_get_original_filenames(self):
        """Test Domain.get_original_filenames."""
        d = cfdm.example_field(0).domain
        self.assertEqual(d.get_original_filenames(), set())

        x = d.coordinate("longitude")
        x._original_filenames(define=["file1.nc", "file3.nc"])
        b = x.bounds
        b._original_filenames(define=["file1.nc", "file4.nc"])

        self.assertEqual(
            d.get_original_filenames(),
            set(
                (
                    cfdm.abspath("file1.nc"),
                    cfdm.abspath("file3.nc"),
                    cfdm.abspath("file4.nc"),
                )
            ),
        )

    def test_Domain_persist(self):
        """Test Domain.persist."""
        f = cfdm.example_field(0)
        cfdm.write(f, tmpfile)
        f = cfdm.read(tmpfile, dask_chunks=3)[0][:2, :2].domain

        on_disk = False
        for k, v in f.coordinate("longitude").data.todict().items():
            if isinstance(v, cfdm.data.abstract.FileArray):
                on_disk = True
                break

        self.assertTrue(on_disk)

        g = f.persist()
        in_memory = False
        for v in g.coordinate("longitude").data.todict().values():
            if isinstance(v, np.ndarray):
                in_memory = True
                break

        self.assertTrue(in_memory)

        # In-place and metdata
        f = cfdm.read(tmpfile)[0].domain
        self.assertIsNone(f.persist(inplace=True))
        in_memory = False
        for v in f.coordinate("longitude").data.todict().values():
            if isinstance(v, np.ndarray):
                in_memory = True
                break

        self.assertTrue(in_memory)


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cfdm.environment()
    print("")
    unittest.main(verbosity=2)



from . import abstract


class CellMeasure(abstract.PropertiesData):
    """A cell measure construct of the CF data model.

    A cell measure construct provides information that is needed about
    the size or shape of the cells and that depends on a subset of the
    domain axis constructs. Cell measure constructs have to be used
    when the size or shape of the cells cannot be deduced from the
    dimension or auxiliary coordinate constructs without special
    knowledge that a generic application cannot be expected to have.

    The cell measure construct consists of a numeric array of the
    metric data which spans a subset of the domain axis constructs,
    and properties to describe the data. The cell measure construct
    specifies a "measure" to indicate which metric of the space it
    supplies, e.g. cell horizontal areas, and must have a units
    property consistent with the measure, e.g. square metres. It is
    assumed that the metric does not depend on axes of the domain
    which are not spanned by the array, along which the values are
    implicitly propagated. CF-netCDF cell measure variables correspond
    to cell measure constructs.

    .. versionadded:: (cfdm) 1.7.0

    """

    def __init__(
        self,
        measure=None,
        properties=None,
        data=None,
        source=None,
        copy=True,
        _use_data=True,
    ):
        """**Initialisation**

        :Parameters:

            measure: `str`, optional
                Set the measure that indicates the metric given by
                the data array.

                The measure may also be set after initialisation with
                the `set_measure` method.

                *Parameter example:*
                  ``measure='area'``

            {{init properties: `dict`, optional}}

                *Parameter example:*
                  ``properties={'units': 'metres 2'}``

            {{init data: data_like, optional}}

            {{init source: optional}}

            {{init copy: `bool`, optional}}

        """
        super().__init__(
            properties=properties,
            source=source,
            data=data,
            copy=copy,
            _use_data=_use_data,
        )

        if source is not None:
            try:
                measure = source.get_measure(None)
            except AttributeError:
                measure = None

        if measure is not None:
            self.set_measure(measure)

    @property
    def construct_type(self):
        """Return a description of the construct type.

        .. versionadded:: (cfdm) 1.7.0

        :Returns:

            `str`
                The construct type.

        **Examples**

        >>> c = {{package}}.{{class}}()
        >>> c.construct_type
        'cell_measure'

        """
        return "cell_measure"

    def del_measure(self, default=ValueError()):
        """Remove the measure.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `get_measure`, `has_measure`, `set_measure`

        :Parameters:

            default: optional
                Return the value of the *default* parameter if the measure
                has not been set.

                {{default Exception}}

        :Returns:

                The removed measure.

        **Examples**

        >>> c = {{package}}.{{class}}()
        >>> c.set_measure('area')
        >>> c.has_measure()
        True
        >>> c.get_measure()
        'area'
        >>> c.del_measure()
        'area'
        >>> c.has_measure()
        False
        >>> print(c.del_measure(None))
        None
        >>> print(c.get_measure(None))
        None

        """
        return self._del_component("measure", default=default)

    def has_measure(self):
        """Whether the measure has been set.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `del_measure`, `get_measure`, `set_measure`

        :Returns:

            `bool`
                True if the measure has been set, otherwise False.

        **Examples**

        >>> c = {{package}}.{{class}}()
        >>> c.set_measure('area')
        >>> c.has_measure()
        True
        >>> c.get_measure()
        'area'
        >>> c.del_measure()
        'area'
        >>> c.has_measure()
        False
        >>> print(c.del_measure(None))
        None
        >>> print(c.get_measure(None))
        None

        """
        return self._has_component("measure")

    def get_measure(self, default=ValueError()):
        """Return the measure.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `del_measure`, `has_measure`, `set_measure`

        :Parameters:

            default: optional
                Return the value of the *default* parameter if the measure
                has not been set.

                {{default Exception}}

        :Returns:

                The value of the measure.

        **Examples**

        >>> c = {{package}}.{{class}}()
        >>> c.set_measure('area')
        >>> c.has_measure()
        True
        >>> c.get_measure()
        'area'
        >>> c.del_measure()
        'area'
        >>> c.has_measure()
        False
        >>> print(c.del_measure(None))
        None
        >>> print(c.get_measure(None))
        None

        """
        return self._get_component("measure", default=default)

    def set_measure(self, measure):
        """Set the measure.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `del_measure`, `get_measure`, `has_measure`

        :Parameters:

            measure: `str`
                The value for the measure.

        :Returns:

             `None`

        **Examples**

        >>> c = {{package}}.{{class}}()
        >>> c.set_measure('area')
        >>> c.has_measure()
        True
        >>> c.get_measure()
        'area'
        >>> c.del_measure()
        'area'
        >>> c.has_measure()
        False
        >>> print(c.del_measure(None))
        None
        >>> print(c.get_measure(None))
        None

        """
        return self._set_component("measure", measure, copy=False)



from . import abstract


class DomainAxis(abstract.Container):
    """A domain axis construct of the CF data model.

    A domain axis construct specifies the number of points along an
    independent axis of the domain. It comprises a positive integer
    representing the size of the axis. In CF-netCDF it is usually
    defined either by a netCDF dimension or by a scalar coordinate
    variable, which implies a domain axis of size one. The field
    construct's data array spans the domain axis constructs of the
    domain, with the optional exception of size one axes, because
    their presence makes no difference to the order of the elements.

    .. versionadded:: (cfdm) 1.7.0

    """

    def __init__(self, size=None, source=None, copy=True):
        """**Initialisation**

        :Parameters:

            size: `int`, optional
                The size of the domain axis.

                *Parameter example:*
                  ``size=192``

                The size may also be set after initialisation with the
                `set_size` method.

            {{init source: optional}}

            {{init copy: `bool`, optional}}

        """
        super().__init__(source=source, copy=copy)

        if source is not None:
            try:
                size = source.get_size(None)
            except AttributeError:
                size = None

        if size is not None:
            self.set_size(size)

    @property
    def construct_type(self):
        """Return a description of the construct type.

        .. versionadded:: (cfdm) 1.7.0

        :Returns:

            `str`
                The construct type.

        **Examples**

        >>> d = {{package}}.{{class}}()
        >>> d.construct_type
        'domain_axis'

        """
        return "domain_axis"

    def del_size(self, default=ValueError()):
        """Remove the size.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `get_size`, `has_size`, `set_size`

        :Parameters:

            default: optional
                Return the value of the *default* parameter if the
                size has not been set.

                {{default Exception}}

        :Returns:

                The removed size.

        **Examples**

        >>> d = {{package}}.{{class}}(size=50)
        >>> d.has_size()
        True
        >>> d.get_size()
        50

        >>> d.set_size(100)
        >>> d.get_size()
        100
        >>> d.del_size()
        100
        >>> d.has_size()
        False
        >>> print(d.del_size(None))
        None

        """
        return self._del_component("size", default=default)

    def has_size(self):
        """Whether the size has been set.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `del_size`, `get_size`, `set_size`

        :Returns:

             `bool`
                True if the size has been set, otherwise False.

        **Examples**

        >>> d = {{package}}.{{class}}(size=50)
        >>> d.has_size()
        True
        >>> d.get_size()
        50

        >>> d.set_size(100)
        >>> d.get_size()
        100
        >>> d.del_size()
        100
        >>> d.has_size()
        False
        >>> print(d.del_size(None))
        None

        """
        return self._has_component("size")

    def get_size(self, default=ValueError()):
        """Return the size.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `del_size`, `has_size`, `set_size`

        :Parameters:

            default: optional
                Return the value of the *default* parameter if the
                size has not been set.

                {{default Exception}}

        :Returns:

                The size.

        **Examples**

        >>> d = {{package}}.{{class}}(size=50)
        >>> d.has_size()
        True
        >>> d.get_size()
        50

        >>> d.set_size(100)
        >>> d.get_size()
        100
        >>> d.del_size()
        100
        >>> d.has_size()
        False
        >>> print(d.del_size(None))
        None

        """
        return self._get_component("size", default=default)

    def set_size(self, size):
        """Set the size.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `del_size`, `get_size`, `has_size`

        :Parameters:

            value: `int`
                The size.

        :Returns:

             `None`

        **Examples**

        >>> d = {{package}}.{{class}}(size=50)
        >>> d.has_size()
        True
        >>> d.get_size()
        50

        >>> d.set_size(100)
        >>> d.get_size()
        100
        >>> d.del_size()
        100
        >>> d.has_size()
        False
        >>> print(d.del_size(None))
        None

        """
        self._set_component("size", size, copy=False)



import numpy as np

from .abstract import MeshSubarray
from .mixin import PointTopology


class PointTopologyFromFacesSubarray(PointTopology, MeshSubarray):
    """A subarray of a point topology array compressed by UGRID faces.

    A subarray describes a unique part of the uncompressed array.

    .. versionadded:: (cfdm) 1.11.0.0

    """

    def _connected_nodes(self, node, node_connectivity, masked):
        """Return nodes that are joined to *node* by face edges.

        The input *node* is included at the start of the returned
        list.

        .. versionadded:: (cfdm) 1.11.0.0

        :Parameters:

            node: `int`
                A node identifier.

            node_connectivity: `numpy.ndarray`
                A UGRID "face_node_connectivity" array.

            masked: `bool`
                Whether or not *node_connectivity* has masked
                elements.

        :Returns:

            `list`
                All nodes that are joined to *node*, including *node*
                itself at the start.

        """
        if masked:
            where = np.ma.where
        else:
            where = np.where

        # Find the faces that contain this node:
        faces = where(node_connectivity == node)[0]

        nodes = []
        nodes_extend = nodes.extend

        # For each face, find which two of its nodes are neighbours to
        # 'node'.
        for face_nodes in node_connectivity[faces]:
            if masked:
                face_nodes = face_nodes.compressed()

            face_nodes = face_nodes.tolist()
            face_nodes.append(face_nodes[0])
            nodes_extend(
                [
                    m
                    for m, n in zip(face_nodes[:-1], face_nodes[1:])
                    if n == node
                ]
            )

        nodes = list(set(nodes))

        # Insert 'node' at the front of the list
        nodes.insert(0, node)

        return nodes



import numpy as np

from . import abstract


class NumpyArray(abstract.Array):
    """A container for a numpy array.

    .. versionadded:: (cfdm) 1.7.0

    """

    def __init__(self, array=None, source=None, copy=True):
        """**Initialisation**

        :Parameters:

            array: `numpy.ndarray`
                The numpy array.

            {{init source: optional}}

            {{init copy: `bool`, optional}}

        """
        super().__init__(source=source, copy=copy)

        if source is not None:
            try:
                array = source._get_component("array", None)
            except AttributeError:
                array = None

        self._set_component("array", array, copy=False)

    @property
    def dtype(self):
        """Data-type of the data elements.

        .. versionadded:: (cfdm) 1.7.0

        **Examples**

        >>> a.dtype
        dtype('float64')
        >>> print(type(a.dtype))
        <type 'numpy.dtype'>

        """
        return self._get_component("array").dtype

    @property
    def shape(self):
        """Tuple of array dimension sizes.

        .. versionadded:: (cfdm) 1.7.0

        **Examples**

        >>> a.shape
        (73, 96)
        >>> a.ndim
        2
        >>> a.size
        7008

        >>> a.shape
        (1, 1, 1)
        >>> a.ndim
        3
        >>> a.size
        1

        >>> a.shape
        ()
        >>> a.ndim
        0
        >>> a.size
        1

        """
        return self._get_component("array").shape

    @property
    def array(self):
        """Return an independent numpy array containing the data.

        .. versionadded:: (cfdm) 1.7.0

        :Returns:

            `numpy.ndarray`
                An independent numpy array of the data.

        **Examples**

        >>> n = a.array
        >>> isinstance(n, numpy.ndarray)
        True

        """
        array = self._get_component("array")

        if not array.ndim and np.ma.isMA(array):
            # This is because np.ma.copy doesn't work for
            # scalar arrays (at the moment, at least)
            ma_array = np.ma.empty((), dtype=array.dtype)
            ma_array[...] = array
            array = ma_array
        else:
            array = array.copy()

        return array

    def copy(self):
        """Return a deep copy of the array.

        ``a.copy() is equivalent to ``copy.deepcopy(a)``.

        Copy-on-write is employed. Therefore, after copying, care must be
        taken when making in-place modifications to attributes of either
        the original or the new copy.

        .. versionadded:: (cfdm) 1.8.7.0

        :Returns:

            `{{class}}`
                The deep copy.

        **Examples**

        >>> b = a.copy()

        """
        klass = self.__class__
        new = klass.__new__(klass)
        new.__dict__ = self.__dict__.copy()
        return new



import datetime
import faulthandler
import os
import unittest

faulthandler.enable()  # to debug seg faults and timeouts

import cfdm


class CellMeasureTest(unittest.TestCase):
    """Unit test for the CellMeasure class."""

    def setUp(self):
        """Preparations called immediately before each test method."""
        # Disable log messages to silence expected warnings
        cfdm.LOG_LEVEL("DISABLE")
        # Note: to enable all messages for given methods, lines or
        # calls (those without a 'verbose' option to do the same)
        # e.g. to debug them, wrap them (for methods, start-to-end
        # internally) as follows:
        #
        # cfdm.LOG_LEVEL('DEBUG')
        # < ... test code ... >
        # cfdm.log_level('DISABLE')

        self.filename = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "test_file.nc"
        )
        f = cfdm.read(self.filename)
        self.assertEqual(len(f), 1)
        self.f = f[0]

    def test_CellMeasure__init__(self):
        """Test the CellMeasure constructor and source keyword."""
        cfdm.CellMeasure(source="qwerty")

    def test_CellMeasure__repr__str__dump_construct_type(self):
        """Test all means of CellMeasure inspection."""
        f = self.f

        for cm in f.cell_measures().values():
            _ = repr(cm)
            _ = str(cm)
            self.assertIsInstance(cm.dump(display=False), str)
            self.assertEqual(cm.construct_type, "cell_measure")

    def test_CellMeasure(self):
        """Test measure access and (un)setting CellMeasure methods."""
        f = self.f.copy()

        cm = f.construct("measure:area")

        self.assertTrue(cm.has_measure())
        self.assertTrue(cm.get_measure(), "area")

        measure = cm.del_measure()
        self.assertFalse(cm.has_measure())
        self.assertIsNone(cm.get_measure(None))
        self.assertIsNone(cm.del_measure(None))

        cm.set_measure(measure)
        self.assertTrue(cm.has_measure())
        self.assertEqual(cm.get_measure(), "area")


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cfdm.environment()
    print("")
    unittest.main(verbosity=2)
